package rione;

import java.io.IOException;

import rescuecore2.components.ComponentLauncher;
import rescuecore2.components.ComponentConnectionException;
import rescuecore2.components.TCPComponentLauncher;
import rescuecore2.connection.ConnectionException;
import rescuecore2.registry.Registry;
import rescuecore2.misc.CommandLineOptions;
import rescuecore2.config.Config;
import rescuecore2.config.ConfigException;
import rescuecore2.Constants;

import rescuecore2.standard.entities.StandardEntityFactory;
import rescuecore2.standard.entities.StandardPropertyFactory;
import rescuecore2.standard.messages.StandardMessageFactory;
import rione.agent.at.AmbulanceCentreAgent;
import rione.agent.at.AmbulanceTeamAgent;
import rione.agent.fb.FireBrigadeAgent;
import rione.agent.fb.FireStationAgent;
import rione.agent.pf.PoliceForceAgent;
import rione.agent.pf.PoliceOfficeAgent;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;

/**
   Launcher for agents. This will launch as many instances of each of the sample agents as possible, all using one connction.
 */
@Deprecated
public final class LaunchAgents {
    private static final Log LOG = LogFactory.getLog(LaunchAgents.class);

    private static final String FIRE_BRIGADE_FLAG = "-fb";
    private static final String FIRE_STATION_FLAG = "-fs";
    private static final String POLICE_FORCE_FLAG = "-pf";
    private static final String POLICE_OFFICE_FLAG = "-po";
    private static final String AMBULANCE_TEAM_FLAG = "-at";
    private static final String AMBULANCE_CENTRE_FLAG = "-ac";
    //private static final String CIVILIAN_FLAG = "-cv";

    private LaunchAgents() {}

    /**
       Launch 'em!
       @param args The following arguments are understood: -p <port>, -h <hostname>, -fb <fire brigades>, -pf <police forces>, -at <ambulance teams>
    */
    public static void main(String[] args) {
        try {
            Registry.SYSTEM_REGISTRY.registerEntityFactory(StandardEntityFactory.INSTANCE);
            Registry.SYSTEM_REGISTRY.registerMessageFactory(StandardMessageFactory.INSTANCE);
            Registry.SYSTEM_REGISTRY.registerPropertyFactory(StandardPropertyFactory.INSTANCE);
            Config config = new Config();
            args = CommandLineOptions.processArgs(args, config);
            int port = config.getIntValue(Constants.KERNEL_PORT_NUMBER_KEY, Constants.DEFAULT_KERNEL_PORT_NUMBER);
            String host = config.getValue(Constants.KERNEL_HOST_NAME_KEY, Constants.DEFAULT_KERNEL_HOST_NAME);
            int fb = -1;
            int fs = -1;
            int pf = -1;
            int po = -1;
            int at = -1;
            int ac = -1;
            // CHECKSTYLE:OFF:ModifiedControlVariable
            for (int i = 0; i < args.length; ++i) {
                if (args[i].equals(FIRE_BRIGADE_FLAG)) {
                    fb = Integer.parseInt(args[++i]);
                }
                else if (args[i].equals(FIRE_STATION_FLAG)) {
                    fs = Integer.parseInt(args[++i]);
                }
                else if (args[i].equals(POLICE_FORCE_FLAG)) {
                    pf = Integer.parseInt(args[++i]);
                }
                else if (args[i].equals(POLICE_OFFICE_FLAG)) {
                    po = Integer.parseInt(args[++i]);
                }
                else if (args[i].equals(AMBULANCE_TEAM_FLAG)) {
                    at = Integer.parseInt(args[++i]);
                }
                else if (args[i].equals(AMBULANCE_CENTRE_FLAG)) {
                    ac = Integer.parseInt(args[++i]);
                }
                else {
                    LOG.warn("Unrecognised option: " + args[i]);
                }
            }
            ComponentLauncher launcer = new TCPComponentLauncher(host, port, config);
            connect(launcer, fb, fs, pf, po, at, ac, config);
        }
        catch (IOException e) {
            LOG.error("Error connecting agents", e);
        }
        catch (ConfigException e) {
            LOG.error("Configuration error", e);
        }
        catch (ConnectionException e) {
            LOG.error("Error connecting agents", e);
        }
        catch (InterruptedException e) {
            LOG.error("Error connecting agents", e);
        }
    }

    private static void connect(ComponentLauncher launcher, int fb, int fs, int pf, int po, int at, int ac, Config config) throws InterruptedException, ConnectionException {
        int i = 0;
        try {
            while (fb-- != 0) {
                LOG.info("Connecting fire brigade " + (i++) + "...");
                launcher.connect(new FireBrigadeAgent());
                LOG.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            LOG.info("failed: " + e.getMessage());
        }
        try {
            while (fs-- != 0) {
                LOG.info("Connecting fire station " + (i++) + "...");
                launcher.connect(new FireStationAgent());
                LOG.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            LOG.info("failed: " + e.getMessage());
        }
        try {
            while (pf-- != 0) {
                LOG.info("Connecting police force " + (i++) + "...");
                launcher.connect(new PoliceForceAgent());
                LOG.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            LOG.info("failed: " + e.getMessage());
        }
        try {
            while (po-- != 0) {
                LOG.info("Connecting police office " + (i++) + "...");
                launcher.connect(new PoliceOfficeAgent());
                LOG.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            LOG.info("failed: " + e.getMessage());
        }
        try {
            while (at-- != 0) {
                LOG.info("Connecting ambulance team " + (i++) + "...");
                launcher.connect(new AmbulanceTeamAgent());
                LOG.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            LOG.info("failed: " + e.getMessage());
        }
        try {
            while (ac-- != 0) {
                LOG.info("Connecting ambulance centre " + (i++) + "...");
                launcher.connect(new AmbulanceCentreAgent());
                LOG.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            LOG.info("failed: " + e.getMessage());
        }
        System.out.println("All agents was connected.");
    }
}