package rione.agent;

import java.util.*;

import rescuecore2.messages.Command;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.MessageManager;
import rione.communication.MessageReader;
import rione.communication.channel.ChannelManager;

public abstract class CommunicationAgent<E extends StandardEntity> extends WorldModelAgent<E> {
	
	//private MessageWriter writer;
	private MessageReader reader;
	protected ChannelManager channelManager;
	protected MessageManager messageManager;
	
	@Override
	protected void initialize() {
		super.initialize();
		reader = new MessageReader(world, config);
		channelManager = new ChannelManager(this, world, config);
		messageManager = new MessageManager(this, world, config);
	}

	protected void subscribe(int...channels) {
		sendSubscribe(time ,channels);
	}
		
	protected void speak(int channel, byte[] data) {
		sendSpeak(time ,channel, data);
	}

	/**
	 * 視界情報から通信情報を作成する
	 */ 
	@Override
	protected void sense(ChangeSet changed) {
		//毎サイクル呼ぶ
		messageManager.init();
		
		//視界情報をキューへ詰め込む
        for (EntityID id : changed.getChangedEntities()) {
        	if (world.time - world.getLastChangedTime(id) < 10) {
        		messageManager.addSendData(world.getEntity(id));
        	}
        }
        
	}

	/**
	 * 通信情報をWorldModelに反映
	 */
	@Override
	protected void hear(Collection<Command> heard) {
		for (Command command : heard) {
			//System.out.println(time+": "+command.toString());
			if (command instanceof AKSpeak) {
				try {
					AKSpeak message = (AKSpeak)command;
					EntityID sender = message.getAgentID();
					StandardEntity entity = world.getEntity(sender);
					if (entity == null) continue;
					if(entity instanceof Civilian) {
						world.addHeardCivilian((Civilian) entity);
						continue;
					}
					reader.open(message);
					do {
						reader.read();
					} while (reader.hasNext());
				} catch (Exception e) {
					 e.printStackTrace();
				} catch (Error e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	@Override
	protected void prepareForAct() {
		super.prepareForAct();
	}

	/**
	 * 指定したチャネルへデータを送信する．
	 * @param c 送信するチャネル．
	 */
	protected void sendmessage(int c) {
        //System.out.println(time+","+c+","+messageManager.getData(c).length);
        speak(c, messageManager.getData(c));
	}


	public ChannelManager getChannelManager() {
		return channelManager;
	}
}
