package rione.agent;
import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.*;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardEntityConstants.BuildingCode;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rione.util.UpdateHashSet;


public abstract class PlatoonAgent<E  extends StandardEntity> extends HumanoidAgent<E> {

	private Map<Integer, EntityID> locationHistory = new HashMap<Integer, EntityID>();

	/** searchBuildingの対象 */
	protected Building searchTarget = null;
	protected StandardEntity pastMe = null;
	/** 火災が起きていないか見回るためのBuilding群 */
	protected UpdateHashSet<Building> searchblds = new UpdateHashSet<Building>() {
		private static final long serialVersionUID = 1L;

		@Override
		public void update() {
			// 探索対象が見えていれば、nullにする
			if (searchTarget != null) {
				if (isVisible(searchTarget) || location().equals(searchTarget) || !(getCommandHistory(time - 1) == StandardMessageURN.AK_MOVE)) {
					searchTarget = null;
				}

				else if (time > 3 && world.getDistance(pastMe, me()) < 3000 && (getCommandHistory(time - 1).equals(StandardMessageURN.AK_MOVE))) {
					StandardEntity se = (world.getEntity(((Human) me()).getPosition()));
					if (se instanceof Road && !reachable.isPassable((Road) se)) {
						removeAll(world.getObjectsInRange(searchTarget, viewDistance * 2));
						remove(searchTarget);
						searchTarget = null;
					}
				}				
				if (world.maybeBurnedOut.contains(searchTarget)) {
					searchTarget = null;
				}
			}
			removeAll(world.maybeBurnedOut);
			// 自分から見えている建物は除外
			for (EntityID id : getVisibleEntitiesID()) {
				Entity e = world.getEntity(id);
				if (e instanceof Building) {
					remove(e);
				}
			}
			// 他のPlatoonAgentが建物にいるとわかった場合も除外
			for (StandardEntity entity : world.getEntitiesOfType(AgentConstants.PLATOONS)) {
				Human human = (Human) entity;
				if (human.isPositionDefined() && human.getPosition(world) instanceof Building) {
					remove(human.getPosition(world));
				}
			}
			// 探索対象のBuilding群が空のとき
			if (isEmpty()) {
				for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BUILDING)) {
					this.add((Building) se);
				}
				for (Iterator<Building> it = iterator(); it.hasNext();) {
					Building building = it.next();
					if ((building.isFierynessDefined() && building.getFierynessEnum() == Fieryness.BURNT_OUT) || building instanceof Refuge || building.isOnFire()) {
						it.remove();
					}
					else if (building.isBrokennessDefined() && building.getBrokenness() > 0) {
						it.remove();
					}
				}
			}

		}
	};

	/** 自分が到達したArea */
	protected UpdateHashSet<Area> visitedArea = new UpdateHashSet<Area>() {
		private static final long serialVersionUID = 1L;
		@Override
		public void update() {
			StandardEntity se = world.getEntity(((Human)me()).getPosition());
			if (se instanceof Area) {
				add((Area) se);
			}
		}
	};

	/**
	 * 他のエージェントが探索済みのエリア
	 */
	protected UpdateHashSet<Area> someoneVisitedArea = new UpdateHashSet<Area>() {
		private static final long serialVersionUID = 1L;
		@Override
		public void update() {
			for (StandardEntity se : world.getEntitiesOfType(AgentConstants.PLATOONS)) {
				Human human = (Human) se;
				//埋まっているエージェントなら無視
				if (human.isBuriednessDefined() && human.getBuriedness() != 0) continue;
				if (human.isPositionDefined()) {
					add((Area) world.getEntity(human.getPosition()));
				}
			}
		}
	};

	private static final double DEFAULT_WOODEN_IGNITION = 47.0f;
	private static final double DEFAULT_STEEL_IGNITION = 47.0f;
	private static final double DEFAULT_CONCRETE_IGNITION = 47.0f;

	private static final String WOODEN_IGNITION_KEY = "resq-fire.wooden_ignition";
	private static final String STEEL_IGNITION_KEY = "resq-fire.steel_ignition";
	private static final String CONCRETE_IGNITION_KEY = "resq-fire.concrete_ignition";
	protected double woodenIgnition;
	protected double steelIgnition;
	protected double concreteIgnition;


	@Override
	protected void initialize() {
		super.initialize();
		pastMe = me();
		woodenIgnition = config.getFloatValue(WOODEN_IGNITION_KEY, DEFAULT_WOODEN_IGNITION);
		steelIgnition = config.getFloatValue(STEEL_IGNITION_KEY, DEFAULT_STEEL_IGNITION);
		concreteIgnition = config.getFloatValue(CONCRETE_IGNITION_KEY, DEFAULT_CONCRETE_IGNITION);

	}

	@Override
	protected void prepareForAct() {
		// world更新
		world.update(changed);
		super.prepareForAct();

		//チャネルの設定
		if (0 < channelManager.getRadioChannelsSize()) {
			int sendChannel = channelManager.getSendChannel();
			sendmessage(sendChannel);
			int[] subscribeChannels = channelManager.getSubscribeChannels();
			for (int subscribeChannel : subscribeChannels) {
				subscribe(subscribeChannel);
			}
		}
		sendmessage(0);

		//順番に注意
		visitedArea.update();
		searchblds.update();
		someoneVisitedArea.update();
	}

	/**
	 * PlatoonAgent共通して行うべき行動
	 */
	@Override
	protected void act() throws ActionCommandException {
		// 建物から出られないときの対策
		try {
			if (getCommandHistory(time-1).equals(StandardMessageURN.AK_MOVE)
					&& getCommandHistory(time-2).equals(StandardMessageURN.AK_MOVE)) {
				if (getLocationHistory(time-1).equals(getLocationHistory(time-2))) {
					if(location() instanceof Building){
						Building bld = (Building)location();
						goOut(bld);
					}
				}
			}
		} catch(NullPointerException e) {}

	}

	/**
	 * 建物の入り口に移動する。
	 * @param bld
	 * @throws ActionCommandException
	 */
	protected void goOut(Building bld) throws ActionCommandException{
		if(world.getEntrance(bld).isEmpty()){
			System.out.println("no Difined Entrance!!");
			move(world.getEntitiesOfType(StandardEntityURN.ROAD));
		}
		else{
			for(Iterator<Road> it=world.getEntrance(bld).iterator(); it.hasNext(); ){
				Road entrance = it.next();
				move(entrance);
			}
		}
	}

	/**
	 * Buildingで火災が起きていないか見回り
	 * @throws ActionCommandException
	 */
	protected void searchBuildings() throws ActionCommandException {
		// 探索対象がないときのみ、探索対象を更新する
		if(searchTarget == null) {
			// 見えていない建物で一番近いところへ行く
			int maxPriority = -1;
			for (Building building : searchblds) {
				int priority = 0;
				//距離考慮
				int distance = world.getDistance(me(), building);
				//見えていない建物だけだから通らないはずだけど一応0除算回避
				if (distance == 0) continue;
				else {
					priority += 100000000 / distance;
				}
				//到達可能考慮
				//if (reachable.getReachableType((Area)building) != ReachableType.False) {
				//	priority += 100000000;
				//}
				if (/*searchTarget == null || */priority > maxPriority) {
					maxPriority = priority;
					searchTarget = building;
				}
			}
		}

		// 探索対象へ移動
		if (me() instanceof PoliceForce) {
			move(searchTarget, CostFunction.pf(reachable, world));
		}
		if (searchTarget != null) {
			move(searchTarget);
		}
	}

	@Override
	protected void afterAct() {
		super.afterAct();
		pastMe = me();
		locationHistory.put(time, location().getID());
	}

	// RandomWalk
	protected void randomWalk() throws ActionCommandException{
		ArrayList<StandardEntity> areas = new ArrayList<StandardEntity>(world.getEntitiesOfType(AgentConstants.AREAS));
		areas.removeAll(world.getIsolationEntities());
		Area area = (Area) areas.get(random.nextInt(areas.size()));
		move(area);
	}

	/**
	 * 近くの避難所に移動する
	 * @throws ActionCommandException
	 */
	protected void moveToRefuge() throws ActionCommandException {
		move(world.getEntitiesOfType(StandardEntityURN.REFUGE));
	}
	/**
	 * 与えられたentityの半径30mの建物の燃えていない率を返す.
	 * @param entity 燃焼度を計算するエンティティ
	 * @return 燃えていない率を返す[0.0~1.0]．ただし，建物が少ない場合は-1.0を返す．
	 */
	public double neighobrHoodsNoFireRate(StandardEntity entity){
		int around = 0;
		int burning = 0;
		int neighobrHoodsDistance = 30*1000; //周辺と判定する距離
		for(StandardEntity e : world.getObjectsInRange(entity, neighobrHoodsDistance)){
			if(e instanceof Building){
				Building b = (Building)e;
				around++;
				if (!b.isFierynessDefined()) {
					continue;
				}
//				String dbgStr = "";
//				dbgStr = b.getFieryness()+","+b.getFierynessEnum()+":";
				if(b.isOnFire()){
					burning++;
//					dbgStr += "isOnFire";
				}else if(b.isFierynessDefined() && b.getFierynessEnum().equals(Fieryness.BURNT_OUT)){
					if(b.isTemperatureDefined() && b.getTemperature()>buildingIgnition(b)){
						burning++;
					}
				}
//				if (b.getFieryness() > 0) System.out.println(dbgStr);
			}
		}
		if (around <= 5) {
			return -1.0;
		}
		//System.out.printf("%9.8f,%d,%d\n",(double)1-((double)burning/around),around,burning);
		return 1.0-((double)burning/around);
	}

	/**
	 * 建物の発火温度を返す
	 * @param bld（Building）
	 * @return 引数の建物の発火温度
	 */
	private double buildingIgnition(Building bld) {
		double temp = 0;
		if(bld.isBuildingCodeDefined()){
			if(bld.getBuildingCodeEnum().equals(BuildingCode.WOOD)){
				temp = woodenIgnition;
			}else if(bld.getBuildingCodeEnum().equals(BuildingCode.STEEL)){
				temp = steelIgnition;
			}else if(bld.getBuildingCodeEnum().equals(BuildingCode.CONCRETE)){
				temp = concreteIgnition;
			}
		}else{
			temp = 50;
		}
		return temp;
	}

	/*
	 * Entityに隣接した(実のところ半径數ｍの範囲内にある)アーキテクトを取ってくる　自己を含まない。
	 */
	protected Set<Entity> getNeighbours(Entity checkedEntity){
		HashSet<Entity> NeighbourEntity = new HashSet<Entity>();
		int r=50;//m単位。
		r*=1000;//mm→mに単位変換。
		for(Entity nears :getAroundsInRange(checkedEntity,r)){
			double rLength;
			rLength=5*1000;
			if(getAroundsInRange(nears,(int)rLength).contains(checkedEntity)){NeighbourEntity.add(nears);}
			NeighbourEntity.remove(checkedEntity);
		}
		return NeighbourEntity;
	}
	/*
	 *周りのエンティティのセットを返す　自分を含まない。
	 */
	protected Set<StandardEntity> getAroundsInRange(Entity e,int range) {
		Set<StandardEntity> Arounds = new HashSet<StandardEntity>();
		for(StandardEntity se : world.getObjectsInRange(e.getID(), range)){
			if(se instanceof Building || se instanceof Road){
				Arounds.add(se);
				Arounds.remove(e);
			}
		}
		return Arounds;
	}
	/*作成中。　Entityの隣のエンティティを持ってくる Buildingは返せるようになった。
	 * Ext=メートル単位。
	 * e=隣のEntityをもってこさせたいEntity(現状Buildingのみ。)(Roadも追加※すごい適当。　11/19 ただし未チェック。)
	 * */
	protected Set<StandardEntity> getNeighbours(StandardEntity e,int Ext){
		printer("StartNeighbour about "+e);
		Set<StandardEntity> Neighbours = new HashSet<StandardEntity>();
		if (e instanceof Building){
			//隣接しているロード。//ついカっとなって追加した//11/19
			for(EntityID tmp2:((Building) e).getNeighbours()){
			Neighbours.add(world.getEntity(tmp2));
			}
			//隣接している建物。
			if(((Building) e).isEdgesDefined()){
				Ext*=1000;
				List<Edge> Edges = new ArrayList<Edge>();
				Edges=((Building) e).getEdges();
				Polygon ExtArea =new Polygon();
				Polygon baseArea = new Polygon();
				int n=1;
				Point2D.Double tmp1=new Point2D.Double();
				Point2D.Double tmp2=new Point2D.Double();
				Point2D.Double tmp3=new Point2D.Double();
				Point2D.Double tmp4=new Point2D.Double();
				Point2D.Double tmp5=new Point2D.Double();
				for(Edge Ee : Edges){baseArea.addPoint(Ee.getStartX(),Ee.getStartY());}
				for(Edge Ee : Edges)
				{
					for(Edge Ee1:Edges){
						if(Ee.getStart().equals(Ee1.getEnd())){
							tmp1.setLocation
							(
									Ee.getStartX()+(n*(Ee.getEndX()-Ee.getStartX()))/Math.hypot((double)Ee.getEndX()-(double)Ee.getStartX() , (double)Ee.getEndY()-(double)Ee.getStartY() )
									,
									Ee.getStartY()+(n*(Ee.getEndY()-Ee.getStartY()))/Math.hypot((double)Ee.getEndX()-(double)Ee.getStartX() , (double)Ee.getEndY()-(double)Ee.getStartY() )
							);
							tmp2.setLocation
							(
									Ee1.getEndX()+(n*(Ee1.getStartX()-Ee1.getEndX()))/Math.hypot((double)Ee1.getStartX()-(double)Ee1.getEndX() , (double)Ee1.getStartY()-(double)Ee1.getEndY() )
									,
									Ee1.getEndY()+(n*(Ee1.getStartY()-Ee1.getEndY()))/Math.hypot((double)Ee1.getStartX()-(double)Ee1.getEndX() , (double)Ee1.getStartY()-(double)Ee1.getEndY() )
							);

							tmp3.setLocation
							(
									(tmp1.x+tmp2.x)/2
									,
									(tmp1.y+tmp2.y)/2
							);
							tmp4.setLocation(Ee.getStartX(), Ee.getStartY());
							if(tmp3.x==tmp4.x&&tmp3.y==tmp4.y){continue;}
							else if(baseArea.contains(tmp3)){
								tmp5.setLocation(
										tmp4.x+(Ext*(tmp4.x-tmp3.x)/Math.hypot(tmp4.x-tmp3.x,tmp4.y-tmp3.y))
										,
										tmp4.y+(Ext*(tmp4.y-tmp3.y)/Math.hypot(tmp4.x-tmp3.x, tmp4.y-tmp3.y)));

							}
							else if (!baseArea.contains(tmp3)){
								tmp5.setLocation(
										tmp4.x+(Ext*(tmp3.x-tmp4.x)/Math.hypot(tmp3.x-tmp4.x,tmp3.y-tmp4.y))
										,
										tmp4.y+(Ext*(tmp3.y-tmp4.y)/Math.hypot(tmp3.x-tmp4.x, tmp3.y-tmp4.y)));
							}
							ExtArea.addPoint((int)tmp5.x, (int)tmp5.y);
						}
					}
				}//ここまでArea拡張。ここまではできてる。
				/*	if(world.getUniform(me())==1){ デバッグメッセージ。
				printer("entityPoints"+ExtArea.npoints);
				for(int returner=0;returner<ExtArea.npoints;returner++){
					System.out.println("base.x:"+baseArea.xpoints[returner]+" base.y:"+baseArea.ypoints[returner]);
				}
				for(int returner=0;returner<ExtArea.npoints;returner++){
					System.out.println("ex.x  :"+ExtArea.xpoints[returner]+" ex.y  :"+ExtArea.ypoints[returner]);
				}
			}
				 */
				for(StandardEntity checker:getAroundsInRange(e,100*1000)){
					if(checker instanceof Building){
						if(((Building)checker).isEdgesDefined()){
							for(Edge Edger:((Building)checker).getEdges()){
								if(ExtArea.contains(Edger.getStartX(),Edger.getStartY())){
									Neighbours.add(checker);//中点も含めるようにしたい。
								}
							}
						}
					}
				}
			}
			else {printer("getEdges Error");}
		}
		else if (e instanceof Road){//11/19追加。
			for(EntityID Neighs:((Road) e).getNeighbours()){
				if(world.getEntity(Neighs) instanceof Road){Neighbours.add(world.getEntity(Neighs));}
				else if (world.getEntity(Neighs) instanceof Building){Neighbours.add(world.getEntity(Neighs));}
			}
		}
		else{printer("getArounds Error");}
		return Neighbours;
	}

	/**
	 * そのサイクルにいた場所のEntityIDを返す。
	 * @param at 時間
	 * @return 場所のEntityID
	 */
	protected EntityID getLocationHistory(int at) {
		return locationHistory.get(at);
	}

	/**
	 * デバッグ用メソッド
	 * @param str
	 */
	public Boolean debugMessengerFlug =false;
	public int debuggerUniform=1;
	protected void debug(String str) {System.out.println(time+" [FB:"+me().getID().getValue()+"] : "+str);	}
	protected void printer(String str){if(debugMessengerFlug && debuggerUniform == world.getUniform(me().getID()))System.out.println(str);}
	protected void printer(int intenger){if(debugMessengerFlug && debuggerUniform == world.getUniform(me().getID()))System.out.println(intenger);}
}
