package rione.agent;

import java.util.*;

import rescuecore2.standard.entities.Area;
import rescuecore2.worldmodel.EntityID;
import rione.object.Street;

public class Route implements Comparable<Route> {

	public final Area end;
	public final Street current; 
	public final List<Street> nexts;
    public double distanceToGoal;
    
    private double cost;
    private final LinkedList<EntityID> route;
    
	
    /**
     * 始点コンストラクタ
     * @param area
     */
	public Route(Street current, Area end, List<EntityID> route, List<Street> neighbors) {
        this.current = current;
		this.end = end;
		this.route = (LinkedList<EntityID>) route;
		this.nexts = neighbors;
	}
    
    /**
     * A*探索 拡張コンストラクタ
     * @param end
     * @param cost
     * @param previous
     */
    private Route(Street current, Area end, LinkedList<EntityID> route, List<Street> neighbors, double cost, Area destination) {
        this.current = current;
    	this.end = end;
    	this.route = route;
        this.nexts = neighbors;
        this.cost = cost;
        distanceToGoal = distance(this.end.getX(), this.end.getY(), destination.getX(), destination.getY());
    }
    
    /**
     * ルートの拡張 A*
     * @param next
     * @param previous
     * @param costFunc
     * @param destination
     * @return
     */
    public Route expandedRoute(Street next, Area previous, CostFunction costFunc, Area destination) {
    	// 次の目標地点、コストの足し算、今までのルート
        return new Route(next, next.getOppositeArea(previous), route, next.getOppositeNeightbor(previous), cost + costFunc.cost(next), destination);
    }
    
    public void setRoute() {
    	route.removeLast();
    	route.addAll(current.getRouteTo(end));
    }
    
    public List<Street> getNexts() {
    	return nexts;
    }
    
    public List<EntityID> getRoute(Area destination) {
    	route.addAll(current.divideRoute(current.getOppositeArea(end), destination));
    	return route;
    }

    public double cost() {
    	return cost;
    }
    
    private double distance(int x1, int y1, int x2, int y2) {
        double dx = x1 - x2;
        double dy = y1 - y2;
        return Math.sqrt((dx * dx) + (dy * dy));
    }
	
	
	@Override
	public int compareTo(Route route) {
		return (int)(this.cost - route.cost);
	}
	
}
