package rione.agent;

import java.util.*;

import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import rione.util.IdSorter;

public abstract class WorldModelAgent<E extends StandardEntity> extends Agent<E> {
	
    private static final int DEFAULT_VIEW_DISTANCE = 30000;
	private static final int DEFAULT_DISTANCE = 30000;

    private static final String VIEW_DISTANCE_KEY = "perception.los.max-distance";
    private static final String MAX_DISTANCE_KEY = "fire.extinguish.max-distance";
 
    /** 視界の範囲[mm] */
    protected int viewDistance;
    
    /** 消火可能距離[mm]　**/
    protected int maxDistance;
 
	protected Random random;

	/** 各エージェントの初期位置 */
	protected Map<EntityID, EntityID> firstPosition;

	@Override
	protected void initialize() {
		super.initialize();
        viewDistance = config.getIntValue(VIEW_DISTANCE_KEY, DEFAULT_VIEW_DISTANCE);
        maxDistance = config.getIntValue(MAX_DISTANCE_KEY, DEFAULT_DISTANCE);
   
        setUniform();	
		random = new Random(getID().getValue());
		firstPosition = new HashMap<EntityID, EntityID>();
	}
	
	@Override
	protected void prepareForAct() {
		super.prepareForAct();
		setFirstPosition();
	}

	/**
	 * エンティティの種類ごとにユニフォームを設定する
	 */
	private void setUniform() {
		Map<EntityID, Integer> id_uniform = new HashMap<EntityID, Integer>();	
		Map<String, Map<Integer, EntityID>> uniformToID = new HashMap<String, Map<Integer, EntityID>>();
		for (StandardEntityURN urn : AgentConstants.UNIFORMABLE_ENTITIES) {
			Map<Integer, EntityID> uniform_id = new HashMap<Integer, EntityID>();
			ArrayList<StandardEntity> entities = new ArrayList<StandardEntity>(world.getEntitiesOfType(urn));
			// EntityID順にソート
			Collections.sort(entities, new IdSorter());
			int i = 0;
			for (StandardEntity entity : entities) {
				id_uniform.put(entity.getID(), i);
				uniform_id.put(i, entity.getID());
				i++;
			}
			uniformToID.put(urn.toString(), uniform_id);
		}
		world.setUniform(id_uniform, uniformToID);
	}
	
	/**
	 * 自身の背番号を返す。
	 * @return 背番号(0~)
	 */
	protected int getUniform() {
		return world.getUniform(getID());
	}
	
	/**
	 * Entityが視界内にあるかどうかを返す
	 * @param id 判定するEntityのID
	 * @return 視界内にあるときtrue，ないときfalse
	 */
	protected boolean isVisible(EntityID id) {
		return changed.getChangedEntities().contains(id);
	}
	
	/**
	 * Entityが視界内にあるかどうかを返す
	 * @param id 判定するEntity
	 * @return 視界内にあるときtrue，ないときfalse
	 */
	protected boolean isVisible(StandardEntity entity) {
		return isVisible(entity.getID());
	}
	
	/**
	 * 視界内にあるEntityの集合を返す．
	 * 自分も含める．
	 * @return 視界内のEntityの集合
	 */
	protected Set<EntityID> getVisibleEntitiesID() {
		return changed.getChangedEntities();
	}
		
	/**
	 * Platoonの初期位置を設定する。
	 * 本当の初期位置ではなく，知覚した最初の位置が登録される．
	 */
	private void setFirstPosition() {
		for (StandardEntity entity : world.getEntitiesOfType(AgentConstants.PLATOONS)) {
			EntityID location = firstPosition.get(entity.getID());
			if (location != null) {
				continue;
			}
			firstPosition.put(entity.getID(), ((Human)entity).getPosition());
		}
	}
}
