package rione.agent.at;

import java.util.*;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.object.ReachableType;
import rione.util.IdSorter;
import rione.util.IntPropertySorter;

public class AmbulanceTeamAgent extends AbstractAmbulanceTeamAgent {

	@Override
	protected void initialize() {
		super.initialize();
		System.out.println(toString() + " was connected. [id=" + getID() + ", uniform=" + getUniform() + "]");
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();
	}

	@Override
	protected void act() throws ActionCommandException {
		try {
			if (getCommandHistory(time - 1).equals(StandardMessageURN.AK_MOVE) && getCommandHistory(time - 2).equals(StandardMessageURN.AK_MOVE)) {
				if (getLocationHistory(time - 1).equals(getLocationHistory(time - 2))) {
					if (location() instanceof Building) {
						Building bld = (Building) location();
						goOut(bld);
					}
				}
			}
		}
		catch (NullPointerException e) {
		}

		carry();

		if (me().getHP() - me().getDamage() * (timestep - time) < 0) {
			moveToRefuge();
		}

		loadOrRescueAround();

		moveToTarget();

		searchBuildings();
		randomWalk();
	}

	/**
	 * loadできるならloadする． RescueできるならRescueする．
	 * 
	 * @throws ActionCommandException
	 */
	protected void loadOrRescueAround() throws ActionCommandException {
		// 視界内のATとHumanを割り当てる
		List<Civilian> cvLoadList = new ArrayList<Civilian>();
		List<Human> humanRescueList = new ArrayList<Human>();
		List<AmbulanceTeam> atList = new ArrayList<AmbulanceTeam>();
		for (EntityID id : getVisibleEntitiesID()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Human) {
				Human hum = (Human) se;
				// 除外
				if (hum.isPositionDefined()) {
					if (world.getEntity(hum.getPosition()) instanceof Refuge || world.getEntity(hum.getPosition()) instanceof Road || world.getEntity(hum.getPosition()) instanceof AmbulanceTeam) {
						continue;
					}
					if (!world.getEntity(hum.getPosition()).equals(location())) {
						continue;
					}
				}
				if (hum.isHPDefined() && hum.getHP() == 0) {
					continue;
				}
				// 救助を行う人
				if (se instanceof AmbulanceTeam && hum.isBuriednessDefined() && hum.getBuriedness() == 0) {
					atList.add((AmbulanceTeam) se);
					continue;
				}
				// 救助される人
				if (hum.isBuriednessDefined()) {
					if (hum.getBuriedness() == 0) {
						if (hum instanceof Civilian) {
							cvLoadList.add((Civilian) hum);
						}
					}
					else {
						humanRescueList.add(hum);
					}
				}
			}
		}
		if (cvLoadList.size() + humanRescueList.size() == 0)
			return;
		// ソート
		// 同じ建物で同じように見えていれば一意に決定される．
		// TODO Distanceの方が通信の情報も使えてていいかも(ただし複雑)
		Collections.sort(cvLoadList, new IdSorter());
		Collections.sort(humanRescueList, new IntPropertySorter(StandardPropertyURN.BURIEDNESS));
		Collections.sort(atList, new IdSorter());

		// atListのインデックス
		int i = 0;
		try {
			// とりあえず運べるやつは運ぶ
			for (; i < cvLoadList.size(); i++) {
				if (atList.get(i).equals(me())) {
					load(cvLoadList.get(i));
				}
			}
			// 早く助かる人間から助ける
			for (Human hum : humanRescueList) {
				// 実験により一人のrescueで1だけburiednessが減る模様．
				// それらしきconfigもないので決め打ち．
				for (int j = 0; j < hum.getBuriedness(); j++) {
					if (atList.get(i).equals(me())) {
						rescue(hum);
					}
					i++;
				}
			}
		}
		catch (IndexOutOfBoundsException e) {
			// 範囲外のインデックスを指定したら終了．
			// ・Loadすべき人数よりATが少ない．
			// ・Rescueに必要な人数よりATが少ない．
		}
	}

	public Set<Area> targetArea = new HashSet<Area>();

	/**
	 * ターゲットに移動する
	 * 
	 * @throws ActionCommandException
	 */
	protected void moveToTarget() throws ActionCommandException {
		final int timeMargin = 5;
		
		targetArea.clear();
		// 100[m]以内のAT
		int nearAT = 0;
		for (StandardEntity se : world.getObjectsInRange(me(), 100 * 1000)) {
			if (se instanceof AmbulanceTeam) {
				nearAT++;
			}
		}
		if (nearAT == 0)
			nearAT++;

		// needRescueHumanから，助かりそうな人がいるAreaだけを選ぶ
		for (EntityID id : world.needRescueHuman) {
			try {
				Human human = (Human) world.getEntity(id);
				if (human == null)
					continue;
				int deathTime = world.getDeathTime(human);
				Building area = (Building) world.getEntity(human.getPosition());
				if (area == null)
					continue;
				if (area.equals(location()))
					continue;

				// 燃えているなら除外
				if (area.isOnFire())
					continue;
				// 
				// if (!world.needRescueArea.contains(area.getID())) continue;
				// 到達のみを受け付ける
				if (!reachable.getReachableType(area).equals(ReachableType.True))
					continue;
				// 助かる可能性があるHumanのみを追加
				if (deathTime < human.getBuriedness() / nearAT + time + timeMargin) {
					targetArea.add(area);
				}
				else if (timestep < deathTime) {
					targetArea.add(area);
				}
			}
			catch (ClassCastException e) {
				continue;
			}
		}
		if (targetArea.size() == 1) {
			// サイズ1ならA*move
			for (Area se : targetArea) {
				move(se);
			}
		}
		else if (targetArea.size() > 1) {
			// サイズ2以上ならダイクストラmove
			move(targetArea);
		}

		// サイズが0ならReachableType.Unknown, Falseも含める
		// あわよくばPFがきてくれることを期待する
		// needRescueHumanから，助かりそうな人がいるAreaだけを選ぶ
		for (EntityID id : world.needRescueHuman) {
			try {
				Human human = (Human) world.getEntity(id);
				if (human == null)
					continue;
				int deathTime = world.getDeathTime(human);
				Building area = (Building) world.getEntity(human.getPosition());
				if (area == null)
					continue;

				// 燃えているなら除外
				if (area.isOnFire())
					continue;
				// 助かる可能性があるHumanのみを追加
				if (deathTime < human.getBuriedness() / nearAT + time + timeMargin) {
					targetArea.add(area);
				}
				else if (timestep < deathTime) {
					targetArea.add(area);
				}
			}
			catch (ClassCastException e) {
				continue;
			}
		}
		if (targetArea.size() == 1) {
			// サイズ1ならA*move
			for (Area se : targetArea) {
				move(se);
			}
		}
		else if (targetArea.size() > 1) {
			// サイズ2以上ならダイクストラmove
			move(targetArea);
		}
	}

	/**
	 * 運ぶ
	 * 
	 * @throws ActionCommandException
	 */
	protected void carry() throws ActionCommandException {
		Civilian board = someoneOnBoard();
		if (board != null) {
			// Am I at a refuge?
			if (location() instanceof Refuge || (board.isHPDefined() && board.getHP() <= 0)) {
				// Unload!
				unload();
			}
			else {
				// Move to a refuge
				moveToRefuge();
			}
		}
	}

	/**
	 * 誰かを乗せているか
	 * 
	 * @return 誰かを乗せていたらtrue
	 */
	public Civilian someoneOnBoard() {
		for (StandardEntity next : world.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian civilian = (Civilian) next;
			if (civilian.isPositionDefined() && civilian.getPosition().equals(getID())) {
				return civilian;
			}
		}
		return null;
	}

	@Override
	public String toString() {
		return "Ri-one ambulance team agent";
	}
}
