package rione.agent.at;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.DebugAgent;
import rione.object.AdvancedWorldModel;
import rione.object.ReachableType;
import rione.viewer.component.AdvancedViewComponent;
import rione.viewer.component.EntityExtension;

/**
 * デバッグ時は実行の構成の起動オプションからこちらを起動する
 * @author utisam
 *
 */
public class DebugAmbulanceTeamAgent extends AmbulanceTeamAgent implements DebugAgent {

	static final boolean showPassableRoad = true;
	static final boolean showReachableBuilding = false;
	static final boolean showBurnedOut = true;
	
	/** 最後に呼び出されたメソッドを表す文字列 */
	private String lastCommand = null;

	@Override
	protected void initialize() {
		super.initialize();
		//ビューアに登録
		AdvancedViewComponent.addTeamAgent(this);
	}

	@Override
	public AdvancedWorldModel getWorld() {
		return world;
	}

	@Override
	public Set<EntityID> getVisibleEntity() {
		return super.getVisibleEntitiesID();
	}

	/*
	 * 自由に設定できるEntityIDと装飾の対応
	 * (非 Javadoc)
	 * @see rione.viewer.component.DebugAgent#customExtension()
	 */
	@Override
	public Map<EntityID, EntityExtension> customExtension() {
		HashMap<EntityID, EntityExtension> exMap = new HashMap<EntityID, EntityExtension>();
		
		for (EntityID id : world.needRescueHuman) {
			exMap.put(id, new EntityExtension(Color.ORANGE, new BasicStroke(3.0f)));
		}
		for (StandardEntity area : targetArea) {
			exMap.put(area.getID(), new EntityExtension(Color.CYAN, new BasicStroke(3.0f)));
		}
		
		if (showPassableRoad) {
			for(StandardEntity one : world.getEntitiesOfType(StandardEntityURN.ROAD)){
				if(reachable.isPassable((Road) one)){
					exMap.put(one.getID(), new EntityExtension(new Color(20, 50, 255, 64), true));
				}
				else {
					exMap.put(one.getID(), new EntityExtension(new Color (255, 20, 20, 64), true));
				}
			}
		}
		if (showReachableBuilding) {
			for(StandardEntity one : world.getEntitiesOfType(StandardEntityURN.BUILDING)){
				if(reachable.getReachableType((Area) one).equals(ReachableType.True)){
					exMap.put(one.getID(), new EntityExtension(new Color(20, 50, 255, 64), true));
				}
				else {
					exMap.put(one.getID(), new EntityExtension(new Color(255, 20, 20, 64), true));
				}
			}
		}
		if (showBurnedOut) {
			for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BUILDING)){
				Building bld = (Building) se;
				if(bld.isFierynessDefined() && bld.getFieryness() == 8){
					exMap.put(bld.getID(), new EntityExtension(Color.WHITE, false));
				}
				else if (world.maybeBurnedOut.contains(bld)) {
					exMap.put(bld.getID(), new EntityExtension(Color.ORANGE, false));
				}
				else {
					exMap.put(bld.getID(), new EntityExtension(Color.MAGENTA, false));
				}
			}
		}
		
		return exMap;
	}
	
	/*
	 * 自由に設定できる描画処理
	 * (非 Javadoc)
	 * @see rione.agent.DebugAgent#customRender(java.awt.Graphics2D, rescuecore2.misc.gui.ScreenTransform, int, int)
	 */
	@Override
	public void customRender(Graphics2D g, ScreenTransform t, int width, int height) {
	}

	@Override
	public String getCommandsCall() {
		if (lastCommand == null) {
			return "";
		}
		return lastCommand;
	}

	///////////////////////////////////////////
	//以下getLastCommandName()用オーバーライド
	//区切り文字は->で統一
	//////////////////////////////////////////
	 
	@Override
	protected void act() throws ActionCommandException {
		lastCommand = "act";
		super.act();
	}
	
	//以下actからの呼び出し
	
	@Override
	protected void carry() throws ActionCommandException {
		lastCommand += "->carry";
		super.carry();
	}

	@Override
	protected void loadOrRescueAround() throws ActionCommandException {
		lastCommand += "->rescueLoadTarget";
		super.loadOrRescueAround();
	}
	
	@Override
	protected void moveToTarget() throws ActionCommandException {
		lastCommand += "->moveToTarget";
		super.moveToTarget();
	}

	@Override
	protected void searchBuildings() throws ActionCommandException {
		lastCommand += "->searchBuildings";
		super.searchBuildings();
	}

	@Override
	protected void randomWalk() throws ActionCommandException {
		lastCommand += "->randomWalk";
		super.randomWalk();
	}
	
	//以下コマンド送信
	
	@Override
	protected void rescue(EntityID target) throws ActionCommandException {
		lastCommand += "->rescue";
		super.rescue(target);
	}

	@Override
	protected void load(EntityID target) throws ActionCommandException {
		lastCommand += "->load";
		super.load(target);
	}
	
	@Override
	protected void unload() throws ActionCommandException {
		lastCommand += "->unload";
		super.unload();
	}
	
	@Override
	protected void moveToRefuge() throws ActionCommandException {
		lastCommand += "->moveToRefuge";
		super.moveToRefuge();
	}
	
	@Override
	protected void move(List<EntityID> path, int x, int y) throws ActionCommandException {
		lastCommand += "->move";
		super.move(path, x, y);
	}
}
