package rione.agent.fb;

import java.awt.geom.Ellipse2D;
import java.util.EnumSet;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.PlatoonAgent;

public abstract class AbstractFireBrigadeAgent extends PlatoonAgent<FireBrigade> {

	private static final int DEFAULT_TANK_CAPACITY = 15000;
	private static final int DEFAULT_POWER = 1000;
	private static final int DEFAULT_REFILL_RATE = 1000;
	private static final int DEFAULT_EXTINGUISHABLE_DISTANCE = 30000;

	//	private static final double DEFAULT_WOODEN_IGNITION = 47.0f;
	//	private static final double DEFAULT_STEEL_IGNITION = 47.0f;
	//	private static final double DEFAULT_CONCRETE_IGNITION = 47.0f;
	private static final String MAX_WATER_KEY = "fire.tank.maximum";
	private static final String MAX_POWER_KEY = "fire.extinguish.max-sum";
	private static final String WATER_REFILL_KEY = "fire.tank.refill-rate";
	private static final String MAX_EXTINGUISHABLE_KEY = "fire.extinguish.max-distance";
	//    private static final String WOODEN_IGNITION_KEY = "resq-fire.wooden_ignition";
	//    private static final String STEEL_IGNITION_KEY = "resq-fire.steel_ignition";
	//    private static final String CONCRETE_IGNITION_KEY = "resq-fire.concrete_ignition";

	/** FBのタンク容量(Configから値を取ってくる) */
	protected int maxTankCapacity;
	/** FBが1サイクルのExtinguishに使える最大水量(Configから値を取ってくる) */
	protected int maxPower;
	/** FBが1サイクルに避難所で給水できる最大水量(Configから値を取ってくる) */
	protected int tankRefillRate;
	/** FBからの消火が届く有効距離(Configから値を取ってくる) 単位[mm] */
	protected int extinguishableDistance;
	// 　下の3つの値はPlatoonAgentに移行
	//    protected double woodenIgnition;
	//    protected double steelIgnition;
	//    protected double concreteIgnition;

	@Override
	protected void initialize() {
		super.initialize();
		maxTankCapacity = config.getIntValue(MAX_WATER_KEY, DEFAULT_TANK_CAPACITY);
		extinguishableDistance = config.getIntValue(MAX_EXTINGUISHABLE_KEY, DEFAULT_EXTINGUISHABLE_DISTANCE);
		maxPower = config.getIntValue(MAX_POWER_KEY, DEFAULT_POWER);

		tankRefillRate = config.getIntValue(WATER_REFILL_KEY, DEFAULT_REFILL_RATE);
	}

	protected void extinguish(EntityID target, int water) throws  ActionCommandException {
		sendExtinguish(time, target, water);
		throw new ActionCommandException(StandardMessageURN.AK_EXTINGUISH);
	}

	protected void extinguish(Building target, int water) throws  ActionCommandException {
		extinguish(target.getID(), water);
	}

	protected void extinguish(EntityID target) throws  ActionCommandException {
		extinguish(target, maxPower);
	}

	protected void extinguish(Building target) throws  ActionCommandException {
		extinguish(target.getID());
	}

	/**
	 * Buildingが消火可能範囲にあるかどうかを返す
	 * @param building 判断するBuilding
	 * @return 消火可能ならtrue
	 * @deprecated 他のBuildingにさえぎられ直接視認できないBuildingは、
	 * 消火後も気づかない可能性がある
	 */
	@Deprecated
	protected boolean isExtinguishable(Building building) {
		Pair<Integer, Integer> position = me().getLocation(world);
		java.awt.geom.Area range = new java.awt.geom.Area(new Ellipse2D.Double(position.first()-viewDistance, position.second()-viewDistance, viewDistance*2, viewDistance*2));
		range.intersect(new java.awt.geom.Area(building.getShape()));
		return !range.isEmpty();
	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}
}
