package rione.agent.fb;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.AgentConstants;
import rione.agent.DebugAgent;
import rione.object.AdvancedWorldModel;
import rione.viewer.component.AdvancedViewComponent;
import rione.viewer.component.EntityExtension;

/**
 * デバッグ時は実行の構成の起動オプションからこちらを起動する
 * @author utisam
 *
 */
public class DebugFireBrigadeAgent extends FireBrigadeAgent implements DebugAgent {

	/** 最後に呼び出されたメソッドを表す文字列 */
	private String lastCommand = null;

	@Override
	protected void initialize() {
		super.initialize();
		//ビューアに登録
		AdvancedViewComponent.addTeamAgent(this);
	}

	@Override
	public AdvancedWorldModel getWorld() {
		return world;
	}

	@Override
	public Set<EntityID> getVisibleEntity() {
		return super.getVisibleEntitiesID();
	}

	/*
	 * 自由に設定できるEntityIDと装飾の対応
	 * (非 Javadoc)
	 * @see rione.viewer.component.DebugAgent#customExtension()
	 */
	@Override
	public Map<EntityID, EntityExtension> customExtension() {
		Map<EntityID,EntityExtension> exMap = new HashMap<EntityID, EntityExtension>();	
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BUILDING)){
			Building bld = (Building) se;
			if(bld.isFierynessDefined() && bld.getFieryness() == 8){
				exMap.put(bld.getID(), new EntityExtension(Color.WHITE, false));
			}
			else if (world.maybeBurnedOut.contains(bld)) {
				exMap.put(bld.getID(), new EntityExtension(Color.ORANGE, false));
			}
			else {
				exMap.put(bld.getID(), new EntityExtension(Color.MAGENTA, false));
			}
		}
		return exMap;
	}

	@Override
	public String getCommandsCall() {
		if (lastCommand == null) {
			return "";
		}
		return lastCommand;
	}
	
	///////////////////////////////////////////
	//以下getLastCommandName()用オーバーライド//
	//////////////////////////////////////////
	 
	@Override
	protected void act() throws ActionCommandException {
		lastCommand = "act";
		super.act();
	}
	
	//以下actからの呼び出し
	
	@Override
	protected void cantMove() throws ActionCommandException {
		lastCommand += "->cantMove";
		super.cantMove();
	}
	
	@Override
	protected void careSelf() throws ActionCommandException {
		lastCommand += "->careSelf";
		super.careSelf();
	}

	@Override
	protected void supplyWater() throws ActionCommandException {
		lastCommand += "->supplyWater";
		super.supplyWater();
	}
	
	@Override
	protected void moveFromfire() throws ActionCommandException {
		lastCommand += "->moveFromfire";
		super.moveFromfire();
	}
	
	@Override
	protected void battleTheFire() throws ActionCommandException {
		lastCommand += "->battleTheFire";
		super.battleTheFire();
	}
	
	@Override
	protected void searchBuildings() throws ActionCommandException {
		lastCommand += "->searchBuildings";
		super.searchBuildings();
	}
	
	@Override
	protected void moveToHadSeen() throws ActionCommandException {
		lastCommand += "moveToHadSeen";
		super.moveToHadSeen();
	}
	
	@Override
	protected void randomWalk() throws ActionCommandException {
		lastCommand += "->randomWalk";
		super.randomWalk();
	}
	
	//以下コマンド送信
	
	@Override
	protected void extinguish(EntityID target, int water) throws  ActionCommandException {
		lastCommand += "->extinguish";
		super.extinguish(target, water);
	}
	
	@Override
	protected void moveToRefuge() throws ActionCommandException {
		lastCommand += "->moveToRefuge";
		super.moveToRefuge();
	}
	
	@Override
	protected void move(List<EntityID> path, int x, int y) throws ActionCommandException {
		lastCommand += "->move";
		super.move(path, x, y);
	}

	@Override
	public void customRender(Graphics2D g, ScreenTransform t, int width,
			int height) {
		for(StandardEntity  se : world.getEntitiesOfType(AgentConstants.HUMANOIDS)){
			if(((Human) se).isPositionDefined()){
				int d=4;
				g.setStroke(new BasicStroke(5.0f));
				if(se instanceof FireBrigade){g.setColor(Color.RED);}
				else if (se instanceof AmbulanceTeam){g.setColor(Color.WHITE);}
				else if (se instanceof PoliceForce){g.setColor(Color.BLUE);}
				else {g.setColor(Color.GREEN);}
		g.drawOval(t.xToScreen(((Human) se).getX())-d, t.yToScreen(((Human) se).getY())-d, 10, 10);
			}
		}// TODO 自動生成されたメソッド・スタブ
		
	}
}
