package rione.agent.fb;

import java.util.*;

import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardPropertyURN;

import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardEntityConstants.BuildingCode;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.messages.StandardMessageURN;
import rione.agent.AgentConstants;
import rione.object.FireArea;
import rione.task.FireBrigadeTask;
import rione.util.DistanceSorter;

public class FireBrigadeAgent extends AbstractFireBrigadeAgent {

	/**
	 * 消火対象のターゲット
	 */
	private Building target = null;
	//ターゲットの履歴。
	private List<Building> targetHistory = new ArrayList<Building>();
	//見たロードを保持。
	private HashSet<EntityID> RoadsLog= new HashSet<EntityID>();
	private Map<StandardEntity, Set<StandardEntity>> NeighborsLog = new HashMap<StandardEntity, Set<StandardEntity>>();
	private Map<StandardEntity, Set<StandardEntity>> NeighBldsLog = new HashMap<StandardEntity, Set<StandardEntity>>();
	private Map<StandardEntity, Set<StandardEntity>> NeighBldsLogX = new HashMap<StandardEntity, Set<StandardEntity>>();
	/**
	 * 給水フラグ
	 */
	private boolean isGettingWater = false;

	/**
	 * 持っているタスク情報（1サイクルごとに初期化）
	 */
	private PriorityQueue<FireBrigadeTask> FBT = new PriorityQueue<FireBrigadeTask>();

	/**
	 * 自分が根城にする（している）避難所
	 */
	private Refuge myRefuge = null;

	/**
	 * 避難所にたどりついたかどうか
	 */
	private boolean reachedRefuge = false;

	/**
	 * 自分が最後に行ったタスク
	 */
	private FireBrigadeTask lastTask = null;

	/**
	 * 自分が消火した火災の周囲の建物
	 */
	private List<Building> aroundExtinguished = new ArrayList<Building>();

	@Override
	protected void initialize() {
		super.initialize();
		debugMessengerFlug=false;
		debuggerUniform=5;
		for(StandardEntity e : world.getEntitiesOfType(AgentConstants.AREAS)){
			setNeighbours(e);
			setNeighBlds(e);
		}
		System.out.println(toString()+" was connected. [id="+getID()+", uniform="+getUniform()+"]");
	}

	@Override
	protected void act() throws ActionCommandException {
		super.act();
		// ターゲットが燃えていないとき、ターゲットを初期化
		unFiredTargetToFormat();
		//動けない時の処理
		cantMove();
		// ダメージをくらったら避難所へ移動する
		careSelf();		
		// 給水処理
		supplyWater();
		//火から退避
		//moveFromfire();
		// ターゲット選択
		selectTarget();
		// ターゲットに移動して消火
		battleTheFire();
		// 火災が起きていないか見回り
		searchBuildings();
		//過去に見に行ったロードを１０サイクル見てないなら見に行く
		moveToHadSeen();
		//ランダムさんまじランダム
		randomWalk();
	}

	@Override
	protected void prepareForAct() {
		// Platoon共通
		super.prepareForAct();

		// タスクを初期化（正常に動作していない様子）
		initializeTasks();

		// 避難所の初期化
		if(myRefuge == null){
			initializeRefuge();
		}
		// 自分の使用しているRefugeの変更
		if(location() instanceof Refuge){
			myRefuge = (Refuge)location();
			reachedRefuge = true;
		}
	}

	/**
	 * 現在使用しているRefugeの更新
	 */
	// TODO prepareForAct()とやってることが若干かぶっているので修正
	protected void initializeRefuge() {
		//最も距離が近いRefugeを設定
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.REFUGE)){
			if(myRefuge == null){
				myRefuge = (Refuge) se;
			}
			else if(world.getDistance(me().getID(), se.getID()) < world.getDistance(me().getID(), myRefuge.getID())){
				myRefuge = (Refuge) se;
			}
		}
	}

	/**
	 * targetが燃えていないとき初期化
	 */
	protected void unFiredTargetToFormat(){
		if (target != null && !target.isOnFire()) {
			target = null;
			lastTask = null;
		}
	}
	/**
	 * targetのBuildingがつながっていない場所のときは無視
	 * @param building
	 * @throws ActionCommandException
	 */
	protected void moveTo(Building building) throws ActionCommandException{
		if(world.isIsolation(building)){
			if (target.equals(building)) {
				target = null;
			}
			return;
		}
		if(building.isOnFire()){
			for(EntityID neighEntityIDs : building.getNeighbours()){
				StandardEntity neighEntitys = world.getEntity(neighEntityIDs);
				if(neighEntitys instanceof Building && !((Building) neighEntitys).isOnFire()){
					move (neighEntityIDs);
				}
				else {
					for( StandardEntity ent : getNeighbours(neighEntitys,5)){
						if(ent instanceof Building &&
								((Building) ent).isFierynessDefined()&&
								!((Building)ent).isOnFire()&&
								world.getDistance(building.getID(), ent.getID())<=extinguishableDistance
						) {
							move(ent.getID());
						}
					}
					move(neighEntityIDs);
				}
			}
		}	
		else move(building);
	}

	/**
	 * 	自分のいるところが燃えてたら逃げる	
	 */
	protected void moveFromfire() throws ActionCommandException {
		if (location() instanceof Building){
			Building mypos = (Building) location();
			if  (mypos.isOnFire()){
				move(world.getEntitiesOfType(StandardEntityURN.ROAD));
			}
		}
	}

	/*  
	 * しばらく見てないロードを見に行く。 
	 * TODO Mapを使ってみる？
	 */
	protected void moveToHadSeen() throws ActionCommandException{
		for(EntityID Rd : getVisibleEntitiesID()){
			if (world.getEntity(Rd) instanceof Road)
			{RoadsLog.add(Rd);}
		}
		EntityID moveTarget=null;
		int tmp1=0,tmp2=0;
		for(EntityID Ro: RoadsLog){
			tmp2=tmp1;
			tmp1=time-world.getLastSeenTime(Ro);
			if(tmp1>tmp2&&tmp1>10){moveTarget=Ro;}
		}
		if(moveTarget!=null){printer("HadSeen");move(moveTarget);}
	}

	/**
	 * 動けない時の処理
	 * @throws ActionCommandException
	 */
	protected void cantMove() throws ActionCommandException{
		// moveしようとしているにも関わらず移動できていないとき
		if (getCommandHistory(time-1) == StandardMessageURN.AK_MOVE
				&& getCommandHistory(time-2) == StandardMessageURN.AK_MOVE
				&& getCommandHistory(time-3) == StandardMessageURN.AK_MOVE ) {
			if (getLocationHistory(time-1) == getLocationHistory(time-2)
					&&getLocationHistory(time-2) == getLocationHistory(time-3) ) {
				if(me().getWater() > 0){
					extinguishHere(); // 手近な火災を消火
				}
				//消火対象が無いならどうにかして出る。
				if(!(location() instanceof Building)){
					List<EntityID> tmp= new ArrayList<EntityID>(((Area) location()).getNeighbours());
					List<EntityID> tmp1= new ArrayList<EntityID>();
					tmp1.add(tmp.get((time%tmp.size())));
					move(tmp1,-1,-1);
				}
			}
		}
	}

	/**
	 * 消火活動
	 */
	protected void battleTheFire() throws ActionCommandException{
		if (target != null) {
			List<FireBrigade> visiblefbs = new ArrayList<FireBrigade>();
			for(EntityID id : getVisibleEntitiesID()){
				if(world.getEntity(id) instanceof FireBrigade){
					visiblefbs.add((FireBrigade)world.getEntity(id));
				}
			}

			Collections.sort(visiblefbs, new DistanceSorter(target, world));


			// ターゲットが消火可能範囲にあるとき、消火する isvisible(target)からtarget.isFierynessDefinedに12/11
			if (world.getPropertyTimeStamp(target.getID(), StandardPropertyURN.FIERYNESS)>=time-1&&target.isFierynessDefined() && world.getDistance(me().getID(), target.getID())<=extinguishableDistance) {
				extinguish(target);
			}

			else if(world.getDistance(me().getID(), target.getID())<=extinguishableDistance && visiblefbs.indexOf(me())<=1){
				moveTo(target);
			}//最寄のだけが見に近づく

			else if (isVisible(target)&& world.getDistance(me().getID(), target.getID())<=extinguishableDistance){
				extinguish(target);
			}
			// 消火可能でないとき、ターゲットにmove
			else {
				moveTo(target);
			}
		}
	}
	/**
	 * 予備放水
	 * @throws ActionCommandException
	 */
	protected void preExtinguish() throws ActionCommandException {
		for (EntityID id : getVisibleEntitiesID()) {
			StandardEntity entity = world.getEntity(id);
			if (entity.getStandardURN().equals(StandardEntityURN.BUILDING)) {
				Building building = (Building)entity;
				if (building.isFierynessDefined() && building.getFierynessEnum().equals(Fieryness.UNBURNT)) {					//System.out.println(getID()+" :pre");
					extinguish(building);
				}
			}
		}
	}

	/**
	 * タンクに水がないとき、満タンになるまで避難所で給水する
	 * @throws ActionCommandException
	 */
	protected void supplyWater() throws ActionCommandException {
		int fbnum = world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE).size();
		int supplyFB; // 避難所で給水する(してもよい)FBの数
		supplyFB = fbnum/2;
		if(me().isWaterDefined() && me().getWater() <= maxTankCapacity) {
			if(location() instanceof Refuge) {
				// 満タンになったら出て行く
				if(me().getWater() >= maxTankCapacity){
					isGettingWater = false;
					return;
				}
				// フラグ変化
				if(refugeFBcount() <= supplyFB) {
					isGettingWater = true;
				}
				else {
					// 自分が出て行くべきかどうか
					EntityID notSupplyId = null; // 出て行くFBのID

					// 現在給水中のFBを取ってくる
					// LinkedListに変更した方が早いかもしれない
					// →間に入れるわけではないのでおｋ
					List<FireBrigade> supplyfbs = new ArrayList<FireBrigade>();

					// 見える範囲のEntityIDを取ってくる
					for(EntityID id : getVisibleEntitiesID()){
						// IDからEntity(実体)を取得
						Entity e = world.getEntity(id);

						// その実体がFireBrigadeで、PositionがRefugeならば、給水中のFBとしてカウント
						if(e instanceof FireBrigade) {
							FireBrigade fb  = (FireBrigade) e;
							if(world.getEntity(fb.getPosition()) instanceof Refuge) {
								supplyfbs.add(fb);
							}
						}
					}

					// 給水中のFBの中で自分の水量が一番多ければ出て行く
					for(FireBrigade fb : supplyfbs){
						if(notSupplyId == null){
							notSupplyId = fb.getID();
						}
						FireBrigade maxWaterFB = (FireBrigade)world.getEntity(notSupplyId);

						// 給水中のFBの中で一番水量の多いFBを給水しない(出て行く)FBとする。
						if(fb.isWaterDefined() && maxWaterFB.isWaterDefined() && fb.getWater() > maxWaterFB.getWater()){
							maxWaterFB = fb;
							notSupplyId = maxWaterFB.getID();
						}
					}

					// 給水中のFBの中で自分が一番水量が多くても、3回以上放水できる量を確保してなければとどまる
					if(notSupplyId != null && me().getID().equals(notSupplyId)
							&& me().getWater() >= (maxTankCapacity<maxPower * 3 ? maxTankCapacity : maxPower * 3)){
						isGettingWater = false;
					}
				}
				// 給水フラグがtrueなら給水する
				if(isGettingWater){
					//moveToRefuge();
					rest();
				}
			}
			// 給水フラグが立っているまたは残量ゼロのとき給水
			else if(me().getWater() == 0 || isGettingWater) {
				isGettingWater = true;
				moveToRefuge();
			}
			// 消火すべき建物がなく、残量が少ないとき給水
			else if(((lastTask != null && visibleFBNum() > lastTask.needChargeNumber / 2) || lastTask == null)
					&& world.getBurningBuildings().isEmpty() && isTankLittle() && refugeFBcount() < supplyFB){
				isGettingWater = true;
				moveToRefuge();
			}
			//水の量が少なくて　かつ　近く(30m)を通ったら給水
			else if (myRefuge != null && world.getDistance(myRefuge.getID(), me().getID())<=30*1000 && me().isWaterDefined() && (double)me().getWater()/maxTankCapacity<=0.3){
				isGettingWater = true;
				moveToRefuge();

			}
			//全体の水量の平均値が一定値を下回った時給水所の近くなら補給にいく。
			else if (checkMeanTankQuantity()){
				isGettingWater = true;
				moveToRefuge();
			}

		}
	}
	/**
	 * 水量の平均値が基準値を下回っているかのチェック＋自分の水量が満タンでないか。+今自分が給水所の近くにいるか
	 */
	private boolean checkMeanTankQuantity() {
		List<FireBrigade> fbsWQ =new ArrayList<FireBrigade>();
		// 現在見えているFBを取ってくる
		for(EntityID id : getVisibleEntitiesID()){
			Entity e = world.getEntity(id);
			if(e instanceof FireBrigade){
				FireBrigade fb  = (FireBrigade)e;
				fbsWQ.add(fb);
			}
		} 
		double sumWQ = -1.0;
		double meanWQ = -1.0;					
		for(FireBrigade fb : fbsWQ) {sumWQ += fb.getWater();}

		if (fbsWQ.size() == 0 ){
			return true;
		}
		else{
			meanWQ = sumWQ/fbsWQ.size();
		}
		// 避難所から10m以内に居る＆周囲のFBの平均水量が3割以下＆自分のタンク水量がMAXでない場合はtrue
		if (world.getDistance(myRefuge.getID(), me().getID())<=10*1000 && me().isWaterDefined() 
				&& meanWQ/maxTankCapacity <=0.3 && me().getWater() != maxTankCapacity){
			return true;
		}
		return false;
	} 

	/**
	 * 給水に行くべき(タンク水量)かどうか判断する
	 * @return 避難所に行く場合true, そうでないときfalse
	 */
	// TODO そもそも他のFBのタンク水量が取れてるか怪しいのでアルゴリズム要検討
	protected boolean isTankLittle() {
		// タンクの中身が最大量あるときは給水の必要なし
		if(me().getWater() >= maxTankCapacity){
			return false;
		}
		EntityID minFB = null; // FBの中で一番水量の少ないFBのIDを保持

		// Worldの中から全てのFBを取得し、それぞれについて調べる
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)){
			FireBrigade fb = (FireBrigade)se;

			// Refugeにいるなら、すでに給水中のFBなので無視
			if(world.getEntity(fb.getPosition()) instanceof Refuge){
				continue;
			}

			// NullPointerExceptionを吐かないよう、nullのときすぐに値を入れておく
			if(minFB == null){
				minFB = fb.getID();
			}
			// 保持しているIDから、タンク水量の一番少ないFBのEntity(実体)を取得
			FireBrigade minTankFB = (FireBrigade)world.getEntity(minFB);

			// 暫定で一番水量の少ないFBよりもタンク水量が少なければ更新
			if(fb.isWaterDefined() && minTankFB.isWaterDefined() && fb.getWater()<minTankFB.getWater()){
				minFB = fb.getID();
			}
		}

		// 自分のIDと、タンク水量の少ないFBのIDが一致すればtrue, そうでなければfalse
		if(me().getID().equals(minFB)){
			return true;
		}
		else{
			return false;
		}
	}

	/**
	 * 50以上ダメージ食らったら避難所へ移動
	 * @throws ActionCommandException
	 */
	protected void careSelf() throws ActionCommandException {
		if (me().getHP() - me().getDamage() * (timestep - time) <= 10){
			moveToRefuge();
		}
	}

	/**
	 * タスクマネージャにある消火済みタスクを報告
	 */
	protected void initializeTasks(){
		FBT.clear();
		// 現在はタスクマネージャを使用しない
		/*
		List<FireBrigadeTask> tasks = new ArrayList<FireBrigadeTask>(taskmanager.getTasks());
		for(FireBrigadeTask check : tasks){
			Building bld = (Building)check.target;
			if(!bld.isOnFire()){
				taskmanager.setFinishTask(check);
			}
		}
		 */
	}

	/**
	 * ターゲットを選択し、targetに格納する
	 */
	protected void selectTarget(){
		Set<Building> targets = new HashSet<Building>(world.getBurningBuildings());
		// 給水中または埋まってないFBを全部取ってくる
		List<FireBrigade> fbs = new ArrayList<FireBrigade>();
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)){
			FireBrigade fb = (FireBrigade)se;
			if( fb.getPosition(world) instanceof Refuge ) { continue; }
			if( fb.isBuriednessDefined() && fb.getBuriedness()>0) { continue; }
			fbs.add(fb);
		}
		// 燃えている火災を全て取ってきてTaskManagerに追加
		if(!targets.isEmpty()) {
			for(Building bld : targets) {
				int chargeNum = calculateChargeNum(bld);
				int priority = calculatePriority(bld);

				// 新しいタスクを生成
				FireBrigadeTask fbt = new FireBrigadeTask(bld,-1,chargeNum);

				// タスクに優先度を設定し、PriorityQueueに入れる
				fbt.setPriority(priority);
				FBT.add(fbt);
			}
		}
		// preExtinguish　(燃えていない建物への予備放水)
		// 温度が取得可能なのは見える建物だけなので、見える建物を取得
		for(EntityID id : getVisibleEntitiesID()){
			if(world.getEntity(id) instanceof Building){
				Building bld = (Building)world.getEntity(id);
				// 避難所とCentre(Center)の場合、また現在燃えている建物の場合は無視
				if(!bld.getFierynessEnum().equals(Fieryness.UNBURNT)
						|| (bld instanceof Refuge)
						|| (bld instanceof AmbulanceCentre)
						|| (bld instanceof PoliceOffice)
						|| (bld instanceof FireStation)
				){continue;}
				int priority = calculatePriority(bld);
				if(priority==Integer.MIN_VALUE){continue;}//温度が閾値以下(発火温度の1/2)無視する
				// タスク生成、Priority設定しPriorityQueueに追加
				FireBrigadeTask fbt = new FireBrigadeTask(bld, -1, 1);
				fbt.setPriority(priority);
				FBT.add(fbt);
			}
		}
		// Priorityの順に上からタスクを取ってくる
		while(!FBT.isEmpty()) {
			FireBrigadeTask fbt = FBT.poll();
			StandardEntity target = world.getEntity(fbt.target);
			// Targetからの距離順にFBをソートし、距離範囲内にいれば取ってくる
			Collections.sort(fbs, new DistanceSorter(target, world));

			// Targetからの距離を順に見て、自分が消火に必要な人数の範囲内にいればターゲットに設定
			if(fbs.indexOf(me()) <= fbt.needChargeNumber){
				lastTask = fbt;
				this.target = (Building) target;
				break;
			}

			// このタスクの建物を消火するFBを考慮(次に優先度の高い建物の消火メンバーに入れない)
			List<FireBrigade> tmp = new ArrayList<FireBrigade>();
			for(int i = fbt.needChargeNumber; i < fbs.size(); i++){
				tmp.add(fbs.get(i));
			}

			// 消火メンバーとなるFBを更新
			fbs.clear();
			fbs.addAll(tmp);
		}
	}
	protected void extinguishHere() throws ActionCommandException {
		//Set<Building> targets = new HashSet<Building>(world.getBurningBuildings());
		Set<Building> targets = getAroundBurning();
		// 燃えている火災を全て取ってきてTaskManagerに追加
		if(!targets.isEmpty()) {
			for(Building bld : targets){
				int chargeNum = calculateChargeNum(bld);
				int priority = calculatePriority(bld);
				FireBrigadeTask fbt = new FireBrigadeTask(bld,-1,chargeNum);
				fbt.setPriority(priority);
				FBT.add(fbt);
			}
		}
		// preExtinguish //12/4 priorityがplusのバージョンにあってないので、Integer.MAX_VALUE/2からculcPriorityのHIGHである6000に変更。 PreExtの閾値を4/5へ※なるだけ水の消費を減らす。ｌ
		for(EntityID id : getVisibleEntitiesID()){
			if(world.getEntity(id) instanceof Building){
				Building bld = (Building)world.getEntity(id);
				if(!bld.isFierynessDefined()) continue;
				if(!bld.getFierynessEnum().equals(Fieryness.UNBURNT) || (bld instanceof Refuge) ||
						(bld instanceof AmbulanceCentre) || (bld instanceof PoliceOffice) || (bld instanceof FireStation)) continue;
				if(world.getDistance(me().getID(), bld.getID()) > extinguishableDistance) continue;
				double temp = 30.0;
				temp = buildingIgnition(bld)*4/5;
				if(bld.isTemperatureDefined() && bld.getTemperature()>(int)temp){
					int priority = 6000; //Integer.MAX_VALUE/2;
					if(bld.isGroundAreaDefined())
						priority += bld.getGroundArea();
					FireBrigadeTask fbt = new FireBrigadeTask(bld, -1, 1);
					fbt.setPriority(priority);
					FBT.add(fbt);
				}
			}
		}
		// Priorityの順に上からタスクを取ってくる
		while( !FBT.isEmpty() ){
			FireBrigadeTask fbt = FBT.poll();
			this.target = (Building)world.getEntity(fbt.target);
			extinguish(target);
		}
	}

	/**
	 * 消火可能範囲または視界範囲内で燃えている建物を取得
	 * @return
	 */
	protected Set<Building> getAroundBurning() {
		Set<Building> blds = new HashSet<Building>();
		for(Building bld : world.getBurningBuildings()){
			if(world.getDistance(me().getID(), bld.getID())
					< (extinguishableDistance<viewDistance?extinguishableDistance:viewDistance)){
				blds.add(bld);
			}
		}
		return blds;
	}

	/**
	 * Refuge内にいるFBの数を返す
	 * @return
	 */
	protected int refugeFBcount(){
		int count = 0;
		for(EntityID id : getVisibleEntitiesID()){
			Entity e = world.getEntity(id);
			if(e instanceof FireBrigade){
				FireBrigade fb = (FireBrigade)e;
				if(world.getEntity(fb.getPosition()) instanceof Refuge){
					count++;
				}
			}
		}
		return count;
	}

	/**
	 * 建物を受け取って、消火に必要な人数を返す
	 * @param bld
	 * @return
	 */
	protected int calculateChargeNum(Building bld) {
		int chargeNum;
		if(bld.isGroundAreaDefined()){
			chargeNum = bld.getGroundArea()/100 + 1;
		}else{
			chargeNum = 5; // マジックナンバー
		}
		if(bld.isFierynessDefined()){
			chargeNum += bld.getFieryness();
		}
		return chargeNum;
	}

	/**
	 * 建物を受け取って、建物のPriorityを返す
	 * @param bld
	 * @return
	 */
	protected int calculatePriority(Building bld) {//　今まで燃えてた中消えても燃えてる建物に囲まれてるなら引き続き消火
		printer("");printer("calcPryStart@"+bld);

		int priority =0;// Integer.MAX_VALUE/2;
		int HIGH=6000,
		MID=3000,
		LOW=1000;//見易さ優先ですが何か。calcPry中の数字弄りたい人がもしいればこれの数字を変更してみたり、calcPryの中のこれらを置き換えてみればいいと思うよ！
		if(bld.isFierynessDefined()){
			//対象が燃えているなら↓
			if(bld.isOnFire()){
				/*sageは差　ageは和。plusが大きいほど優先度は上がる・・・はず*/
				//建物の材質によって優先度上昇※
				if(bld.isBuildingCodeDefined()){ 
					priority += buildingIgnition(bld)*100;
				}
				//近隣AreaCheckここから
				//周りに建物が無いなら(ほぼ)無視する。
				if(getNeighBlds(bld).isEmpty()){
					printer("noThings arounds "+bld);
					priority -=Integer.MAX_VALUE/2;
				}
				//周りに建物があるなら～
				else{printer("NeighID=");
				/*温度低い方優先 sage*///Buildingに囲まれているほど優先度が上がる・・・はずなんだけどな・・・
				if(bld.isTemperatureDefined()){priority -= bld.getTemperature()*50*getNeighBlds(bld).size()/*参考:発火温度50くらい？*/;}
				//対象のbuildingの周りの建物を全て取ってきて検査
				for (Entity neighbourEntity :getNeighBlds(bld)){
					printer(" "+neighbourEntity);//隣接しているBuildingを表示
					/*建物をBuilding型に*/					Building NeighbourBld = (Building) neighbourEntity;	
					/*隣接建物が燃えてないなら優先度age*/		if (!NeighbourBld.isOnFire()) {priority +=HIGH;}
					/*周りよりもすごく燃えてると優先度sage*/     //else if (bld.getFieryness()>NeighbourBld.getFieryness()*5/*どのくらい差があるか*/){priority-=MID; }
					/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					if(NeighbourBld.isFierynessDefined()){
						/*隣の建物が燃え尽きていたら優先度sage*/
						if (NeighbourBld.getFierynessEnum().equals(Fieryness.BURNT_OUT)){priority -=LOW;}
						else if (NeighbourBld.getFierynessEnum().equals(Fieryness.MINOR_DAMAGE)){priority+=LOW;}
						else if (NeighbourBld.getFierynessEnum().equals(Fieryness.SEVERE_DAMAGE)){priority+=LOW;}
						else if (NeighbourBld.getFierynessEnum().equals(Fieryness.MODERATE_DAMAGE)){priority+=LOW;}
						else if (NeighbourBld.getFierynessEnum().equals(Fieryness.WATER_DAMAGE)){priority+=MID;}
					}
				}printer(" ");
				}//近隣AreaCheckここまで

				//消火地域がtargetと同じ　建物なら優先度age
				List<FireArea> FireAreas = world.getFireAreas();
				for (FireArea it : FireAreas){
					if(it.getBuildings().contains(bld)){
						if(it.getBuildings().contains(target)){priority +=LOW;}
						//修正により使用してみる。
						priority-=( it.getBurntOutBuildings().size()*1000)/it.getBuildings().size();
						priority +=( it.getBurningBuildings().size()*1000)/it.getBuildings().size();
						priority -=(it.getBuildings().size())*1000;
					}
				}

				if(location() instanceof Refuge){//reachedRefuge
					priority -= (world.getDistance(bld.getID(), myRefuge.getID())*2)  
					*(1+me().getDamage()/5000 - me().getWater()/maxTankCapacity);
				}else{
					priority -= world.getDistance(bld.getID(),me().getID())*2;
				}

				/*target変更するとしばらく優先度うｐ どうしてもふら付きを抑えたいならONにしてもいいかも。柔軟度さがるけど。
		if(targetHistory.isEmpty()){targetHistory.add(target);}
		if (targetHistory.getLast()!=target){changeWeight=3;}//何サイクル間優先度を上げておくか
		if(changeWeight>0 && bld ==target ){minus*=1/changeWeight; changeWeight--;}
		targetHistory.add(target);
				 */

				//消化可能範囲にあれば優先度age
				if ( world.getDistance(me().getID(), bld.getID())<=extinguishableDistance ){priority +=HIGH*2;}
			}



			//対象が燃えていないなら↓//つまりPreExtingush用。
			else if(!bld.isOnFire()){
				// Buildingの発火温度を取得
				double 	temp = buildingIgnition(bld);
				// Buildingの発火温度の半分の温度まで達していれば、予備放水をするべき //preExtingish//Integer.MAX_VALUE/2>>6000
				if(bld.isTemperatureDefined() && bld.getTemperature() > (int)temp/2) {
					priority =0;
					// Priorityの基準値
					priority = HIGH;//Integer.MAX_VALUE / 2;

					//仮実装。　消した直後の建物は引き続き消火(という名の優先度上昇。
					if(
							bld.getTemperature()/temp>0.9 
							&& world.getDistance(me().getID(), bld.getID()) <= extinguishableDistance
							&& neighboursFirerinessChecker(bld,1)
					){priority +=HIGH;}
					//仮実装ここまで
					//PreExtにおいてCivilianのいる建物に優先。
					for(Entity Civ :world.getEntitiesOfType(StandardEntityURN.CIVILIAN)){
						if(bld==world.getEntity(((Civilian)Civ).getPosition())){
							priority+=LOW;
						}
					}
					// 面積の広いBuildingを優先させる
					if(bld.isGroundAreaDefined()){
						priority += bld.getGroundArea();
					}
					//囲んでる建物が多いとき優先させる
					//周りが何も無いかロードだけなら無視させる。
					if (getNeighBlds(bld).isEmpty()){priority=Integer.MIN_VALUE;}
					else{priority +=getNeighBlds(bld).size()*1000;}
					//無視させるここまで
				}
				else{priority = Integer.MIN_VALUE;}
			}
		}//燃え具合が確認できるなら。↑まで。
		else { priority=Integer.MIN_VALUE;}
		if (priority==Integer.MIN_VALUE){printer("calcPryEnd@"+bld+" priority=Ignored");}
		else{printer("calcPryEnd@"+bld+" priority=" + priority);}
		
		if (world.determinatedBurnedOut.size() > 10) {
			if (world.determinatedBurnedOut.contains(bld)) priority /= 2;
			if (world.maybeBurnedOut.contains(bld)) priority /= 2;
		}
		
		return priority;
	}

	/**
	 * 建物の発火温度を返す
	 * @param bld（Building）
	 * @return 引数の建物の発火温度
	 */
	protected double buildingIgnition(Building bld) {
		if(bld.isBuildingCodeDefined()){
			if(bld.getBuildingCodeEnum().equals(BuildingCode.WOOD)){
				return  woodenIgnition;
			}else if(bld.getBuildingCodeEnum().equals(BuildingCode.STEEL)){
				return  steelIgnition;
			}else if(bld.getBuildingCodeEnum().equals(BuildingCode.CONCRETE)){
				return  concreteIgnition;
			}
		}
		return  50.0;
	}

	/**
	 * 自分の視界内から見えるFBの数を返す
	 * @return 自分から見えるFBの数
	 */
	protected int visibleFBNum() {
		int count = 0;
		for(EntityID id : getVisibleEntitiesID()){
			if(id.equals(me().getID())) continue;
			if(world.getEntity(id) instanceof FireBrigade){
				count++;
			}
		}
		return count;
	}

	/**
	 * Entityから半径range[mm]以内にある建物を取ってくる
	 * @param e
	 * @param range
	 * @return
	 */
	protected Set<Building> getAroundBuildings(Entity e,int range) {
		Set<Building> blds = new HashSet<Building>();
		for(StandardEntity se : world.getObjectsInRange(e.getID(), range)){
			if(se instanceof Building){
				blds.add((Building)se);
			}
		}
		return blds;
	}

	@Override
	public String toString() {
		return "Ri-one fire brigade agent";
	}

	/*
	 * 対象の建物がの具合がどうかを返す 燃えてる数ー燃え尽きた数＞numbersならtrue
	 */
	protected boolean neighboursFirerinessChecker(StandardEntity bld,int numbers){
		int count=0;
		for(Entity e:getNeighbours(bld,5)){
			if(e instanceof Building && ((Building) e).isFierynessDefined() && ((Building) e).isOnFire()){count++;}
			if(e instanceof Building && ((Building) e).isFierynessDefined() && ((Building) e).getFierynessEnum().equals(Fieryness.BURNT_OUT)){count--;}
		}
		if (count>=numbers){return true;}
		else {return false;}
	}





	//軽量化策※メモリ食い
	@Override
	protected Set<StandardEntity> getNeighbours(StandardEntity e,int Ext){
		if (Ext == 5) return NeighborsLog.get(e);
		return super.getNeighbours(e,Ext);
	}
	protected void setNeighbours(StandardEntity e){
		NeighborsLog.put(e, super.getNeighbours(e, 5));
	}

	protected Set<StandardEntity> getNeighBlds (StandardEntity e){
		return NeighBldsLog.get(e);
	}
	protected void setNeighBlds( StandardEntity e ){
		Set <StandardEntity> bldSets = new HashSet<StandardEntity>(16, (float) 0.9);
		for(StandardEntity ent : getNeighbours(e,5)){
			StandardEntity se = ent;
			if(	!(se instanceof Refuge) 
					& !(se instanceof Road)
					& !(se instanceof AmbulanceCentre) 
					& !(se instanceof PoliceOffice) 
					& !(se instanceof FireStation)
					& (se instanceof Building)) {
				bldSets.add(ent);
			}
		}
		NeighBldsLog.put(e, bldSets); 
		bldSets.clear();
		for(StandardEntity ent : getNeighbours(e,20)){
			StandardEntity se = ent;
			if(	!(se instanceof Refuge) 
					& !(se instanceof Road)
					& !(se instanceof AmbulanceCentre) 
					& !(se instanceof PoliceOffice) 
					& !(se instanceof FireStation)
					& (se instanceof Building)) {
				bldSets.add(ent);
			}
		}
		NeighBldsLogX.put(e, bldSets);
	}
}
