package rione.agent.pf;

import java.awt.geom.Ellipse2D;
import java.util.*;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.PlatoonAgent;

public abstract class AbstractPoliceForceAgent extends
		PlatoonAgent<PoliceForce> {

	private static final int DEFAULT_REPAIR_RATE = 10;
	private static final int DEFAULT_DISTANCE = 10000;

	private static final String DISTANCE_KEY = "clear.repair.distance";
	private static final String REPAIR_RATE_KEY = "clear.repair.rate";

	// protected PoliceForceTaskManager taskManager;
	protected Set<Blockade> clearedBlockade = new HashSet<Blockade>();
	/** 除去可能な距離 */
	protected int repairDistance;
	protected int repairRate;

	@Override
	protected void initialize() {
		super.initialize();
		repairDistance = config.getIntValue(DISTANCE_KEY, DEFAULT_REPAIR_RATE);
		repairRate = config.getIntValue(REPAIR_RATE_KEY, DEFAULT_DISTANCE);
		// taskManager = new PoliceForceTaskManager(world);
	}

	/**
	 * EntityIDを渡したClear
	 * 
	 * @param target
	 *            Clearする対象のID
	 * @throws ActionCommandException
	 */
	protected void clear(EntityID block) throws ActionCommandException {
		sendClear(time, block);
		throw new ActionCommandException(StandardMessageURN.AK_CLEAR);
	}

	/**
	 * Blockadeを渡したClear
	 * 
	 * @param block
	 *            Clearする対象
	 * @throws ActionCommandException
	 */
	protected void clear(Blockade block) throws ActionCommandException {
		clear(block.getID());
	}

	/**
	 * ある閉塞が啓開できる距離にあるかどうかを判断する
	 * 
	 * @param block
	 *            判断する閉塞
	 * @return 閉塞が啓開できる距離にあるとき true
	 */
	protected boolean isClearable(Blockade block) {
		Pair<Integer, Integer> position = me().getLocation(world);
		java.awt.geom.Area range = new java.awt.geom.Area(new Ellipse2D.Double(
				position.first() - repairDistance, position.second() - repairDistance,
				repairDistance * 2, repairDistance * 2));
		range.intersect(new java.awt.geom.Area(block.getShape()));
		return !range.isEmpty();
	}

	/**
	 * ある閉塞が指定したPoliceForceAgentが 啓開できる距離にあるかどうかを判断する
	 * 
	 * @param pf
	 *            指定したPoliceForceAgent
	 * @param block
	 *            判断する閉塞
	 * @return 閉塞が啓開できる距離にあるとき true
	 */
	protected boolean isClearable(PoliceForce pf, Blockade block) {
		Pair<Integer, Integer> position = pf.getLocation(world);
		java.awt.geom.Area range = new java.awt.geom.Area(new Ellipse2D.Double(
				position.first() - repairDistance, position.second() - repairDistance,
				repairDistance * 2, repairDistance * 2));
		range.intersect(new java.awt.geom.Area(block.getShape()));
		return !range.isEmpty();
	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.POLICE_FORCE);
	}
}
