package rione.agent.pf;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.Building;
import rescuecore2.worldmodel.EntityID;
import rione.agent.DebugAgent;
import rione.object.AdvancedWorldModel;
import rione.viewer.AdvancedViewer;
import rione.viewer.component.AdvancedViewComponent;
import rione.viewer.component.EntityExtension;

/**
 * デバッグ時は実行の構成の起動オプションからこちらを起動する
 * @author utisam
 *
 */
public class DebugPoliceForceAgent extends PoliceForceAgent implements DebugAgent {

	/** 最後に呼び出されたメソッドを表す文字列 */
	private String lastCommand = null;

	@Override
	protected void initialize() {
		super.initialize();
		//ビューアに登録
		AdvancedViewComponent.addTeamAgent(this);
	}

	@Override
	public AdvancedWorldModel getWorld() {
		return world;
	}

	@Override
	public Set<EntityID> getVisibleEntity() {
		return super.getVisibleEntitiesID();
	}

	/*
	 * 自由に設定できるEntityIDと装飾の対応
	 * (非 Javadoc)
	 * @see rione.viewer.component.DebugAgent#customExtension()
	 */
	@Override
	public Map<EntityID, EntityExtension> customExtension() {
		Map<EntityID, EntityExtension> exMap = new HashMap<EntityID, EntityExtension>();
		
		try {
			exMap.put(targetTask.target, new EntityExtension(Color.CYAN, new BasicStroke(3.0f)));
			
			for (EntityID id : ((Building)location()).getNeighbours())
				exMap.put(id, new EntityExtension(Color.PINK, new BasicStroke(3.0f)));
		}catch (NullPointerException e) {e.printStackTrace();}
		catch (java.lang.ClassCastException e){}
		
		
		
		return exMap;
	}

	@Override
	public String getCommandsCall() {
		if (lastCommand == null) {
			return "";
		}
		return lastCommand;
	}
	
	@Override
	public void customRender(Graphics2D g, ScreenTransform t, int width, int height) {
		g.setColor(Color.YELLOW);
		g.drawString(AdvancedViewer.createDetailString(targetTask.target), 16, 32);
	}
	
	///////////////////////////////////////////
	//以下getLastCommandName()用オーバーライド//
	//////////////////////////////////////////
	 
	@Override
	protected void act() throws ActionCommandException {
		lastCommand = "act";
		super.act();
	}
	
	//以下actからの呼び出し
	
	@Override
	protected void goOut(Building bld) throws ActionCommandException {
		lastCommand += "->goOut";
		super.goOut(bld);
	}
	
	@Override
	protected void clearAround() throws ActionCommandException {
		lastCommand += "->clearAround";
		super.clearAround();
	}
	
	@Override
	protected void moveToTarget() throws ActionCommandException {
		lastCommand += "->moveToTarget";
		super.moveToTarget();
	}

	@Override
	protected void searchBuildings() throws ActionCommandException {
		lastCommand += "->searchBuildings";
		super.searchBuildings();
	}

	@Override
	protected void randomWalk() throws ActionCommandException {
		lastCommand += "->randomWalk";
		super.randomWalk();
	}
	
	//以下コマンド送信
	
	@Override
	protected void clear(EntityID block) throws ActionCommandException {
		lastCommand += "->clear";
		super.clear(block);
	}
	
	@Override
	protected void moveToRefuge() throws ActionCommandException {
		lastCommand += "->moveToRefuge";
		super.moveToRefuge();
	}
	
	@Override
	protected void move(List<EntityID> path, int x, int y) throws ActionCommandException {
		lastCommand += "->move";
		super.move(path, x, y);
	}
}
