package rione.agent.pf;

import java.util.*;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.AgentConstants;
import rione.agent.CostFunction;
import rione.object.ReachableType;
import rione.task.Task;
import rione.task.PoliceForceTask;
import rione.util.DistanceSorter;
import rione.util.IdSorter;
import rione.util.UpdateHashSet;

public class PoliceForceAgent extends AbstractPoliceForceAgent {

	/** 無視して良い瓦礫 */
	protected UpdateHashSet<Blockade> unnecessaryBlockades = new UpdateHashSet<Blockade>() {
		private static final long serialVersionUID = 1L;

		@Override
		public void update() {
			for (StandardEntity e : world.getEntitiesOfType(StandardEntityURN.BLOCKADE)) {
				if (!(e instanceof Blockade))
					continue;
				Blockade b = (Blockade) e;
				if (isForgettableBlock(b)) {
					add(b);
				}
				if (isNeedBlock(b)) {
					remove(b);
				}
			}
		}

		/**
		 * 情報が追加され，必要となったかどうか
		 * 
		 * @param b
		 * @return 必要ならtrue
		 */
		private boolean isNeedBlock(Blockade b) {
			// 内部に人が居るときは必要
			if (world.isContainsHuman(b)) {
				return true;
			}
			// 属しているエリアの中心を覆っていたら必要
			if (b.isPositionDefined() && b.isXDefined() && b.isYDefined()) {
				Area area = (Area) world.getEntity(b.getPosition());
				if (area != null && area.getShape().contains(b.getX(), b.getY())) {
					return true;
				}
			}
			return false;
		}

		private boolean isForgettableBlock(Blockade block) {
			// 頂点がひとつ(通信で得た)かどうか
			if (block.isApexesDefined() && block.getApexes().length <= 1) {
				// 通信で得た瓦礫
				return false;
			}
			else {
				// 除去コストが小さければ無視
				// Kobeだと小さな入り口Roadが無視されてしまう
				// 近くに重要な建物とかHumanがなければ外しても良い
				// TODO ↑判定できるようになればなおよい
				if (block.isRepairCostDefined() &&
					block.getRepairCost() < 1) {
					return true;
				}
				if (block.isPositionDefined()) {
					// 矛盾したblockadeは無視
					StandardEntity entity = world.getEntity(block.getPosition());
					if (entity instanceof Road) {
						Road road = (Road) entity;
						if (road.isBlockadesDefined()) {
							if (!road.getBlockades().contains(block.getID())) {
								return true;
							}
						}
					}
					// 道を塞ぎきっていないものは無視
					if (reachable.isIgnorableBlockad(block)) {
						return true;
					}
				}
			}
			return false;
		}
	};

	/** 到達可能な(いずれかのPFが到達した)Refuge */
	protected UpdateHashSet<Refuge> passedRefuge = new UpdateHashSet<Refuge>() {
		private static final long serialVersionUID = 1L;

		@Override
		public void update() {
			// いずれかのPFが探索したRefugeの場所を保存しておく
			for (StandardEntity entity : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
				PoliceForce pf = (PoliceForce) entity;
				if (!pf.isPositionDefined())
					continue;
				StandardEntity area = world.getEntity(pf.getPosition());
				if (area instanceof Refuge) {
					add((Refuge) area);
				}
			}
		}
	};

	/** ターゲットのタスク */
	protected PoliceForceTask targetTask = null;

	@Override
	protected void initialize() {
		super.initialize();
		// 接続表示
		System.out.println(toString() + " was connected. [id=" + getID() + ", uniform=" + getUniform() + "]");

		// 最初に行くターゲット(現在はRefuge)を設定する
		List<StandardEntity> pfs = new ArrayList<StandardEntity>(world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE));
		List<StandardEntity> refuges = new ArrayList<StandardEntity>(world.getEntitiesOfType(StandardEntityURN.REFUGE));
		if (refuges.size() != 0) {
			Collections.sort(pfs, new IdSorter());
			for (int i = 0; i < pfs.size(); i++) {
				// ID順にソートしてi番目なら，(i % refuge数)を割り当てる
				if (me().getID().equals(pfs.get(i).getID())) {
					targetTask = new PoliceForceTask(refuges.get(i % refuges.size()), Task.high * 1000, (i % 2 == 0) ? PoliceForceTask.NORMAL_FUNCTION : PoliceForceTask.PF_FUNCTION);
				}
			}
		}
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();
		unnecessaryBlockades.update();
		passedRefuge.update();

		if (checkFinishTarget()) {
			targetTask = selectTask();
		}
	}

	private PoliceForceTask selectTask() {
		PoliceForceTask maxPriorityTask = null;
		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.HUMANOIDS)) {
			int priority = getHumanPriority((Human) se);
			if (priority > Task.zero && (maxPriorityTask == null || maxPriorityTask.priority < priority)) {
				maxPriorityTask = new PoliceForceTask(se, priority, getApproach(se));
			}
		}
		if (maxPriorityTask != null && maxPriorityTask.priority > Task.middle)
			return maxPriorityTask;
		if (getID().getValue() % 2 == 0) {
			for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BUILDING)) {
				int priority = getAreaPriority((Area) se);
				if (priority > Task.zero && (maxPriorityTask == null || maxPriorityTask.priority < priority)) {
					maxPriorityTask = new PoliceForceTask(se, priority, getApproach(se));
				}
			}
		}
		else {
			for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BLOCKADE)) {
				int priority = getBlockadePriority((Blockade) se);
				if (priority > Task.zero && (maxPriorityTask == null || maxPriorityTask.priority < priority)) {
					maxPriorityTask = new PoliceForceTask(se, priority, getApproach(se));
				}
			}
		}
		return maxPriorityTask;
	}

	/**
	 * タスクの対処方法を決定します，
	 * 
	 * @return
	 */
	private int getApproach(StandardEntity se) {
		// TODO おそらくCostfunction.pfからいじらないとだめ．
		if (se instanceof Area) {
			// 誰かが到達したことがあるなら最短を開ける
			// それ以外なら最短で開ける
			return (someoneVisitedArea.contains(se)) ? PoliceForceTask.PF_FUNCTION : PoliceForceTask.NORMAL_FUNCTION;
		}
		else {
			// 適当
			return ((getUniform() + time) % 3 == 0) ? PoliceForceTask.PF_FUNCTION : PoliceForceTask.NORMAL_FUNCTION;
		}
	}

	private int getBlockadePriority(Blockade block) {
		// まず決定的要素を決める
		// 無視
		if (unnecessaryBlockades.contains(block)) {
			return Task.minus;
		}

		// double result = Task.middle;
		// if (isMytask(block)) {
		// // 自分でなくても良い
		// result = Task.low;
		// }
		double result = Task.low;
		if (world.isContainsHuman(block)) {
			// 中に人が居るか
			result = Task.high;
		}

		// 周辺のエージェント考慮
		result *= aroundAgentsLate(block, (viewDistance * time / timestep) >> 2);// 時間が進むごとに広げてみる
		// 周辺の火災考慮
		result *= aroundFireLate(block);// 時間が進むごとに広げてみる
		// 距離考慮
		result *= distanceLate(block);

		return (int) result;
	}

	private double aroundFireLate(StandardEntity se) {
		double late = neighobrHoodsNoFireRate(se);
		double result = 0.0;

		if (0.9 < late) {
			// 全く燃えていないであろう場合
			result = 0.75;
		}
		else if (0.25 < late) {
			// 燃え始め~中盤
			result = late * 4.0;
		}
		else if (0.0 < late) {
			// 燃えすぎ(燃えている,燃え尽きた建物が小さいとき)
			result = 0.4;
		}
		else {
			// 建物少なすぎ
			result = 0.65;
		}

		return Math.min(1.0, result);
	}

	/**
	 * StandardEntity周辺のFBとATの数と優先度の対応
	 * 
	 * @param se
	 *                対象のStandardEntity
	 * @param range
	 *                広げる範囲
	 * @return 優先度(0.0〜1.0)
	 */
	protected double aroundAgentsLate(StandardEntity se, int range) {
		double result = 0.0;
		boolean ATFlg = false;
		boolean CVFlg = false;

		for (StandardEntity rse : world.getObjectsInRange(se, range)) {
			if (rse instanceof Human) {
				// 2人以上いたら十分多いと思う．
				result += humanTypeLate((Human) rse) / 2;
				if (rse instanceof Civilian) {
					CVFlg = true;
				}
				else if (rse instanceof AmbulanceTeam) {
					ATFlg = true;
				}
			}
		}
		// ATがよく市民の居る建物の前で引っかかるので応急措置
		return Math.min(1.0, result) * ((ATFlg && CVFlg) ? 1.5 : 1.0);
	}

	private int getAreaPriority(Area area) {
		// 今いる場所は無視
		if (area.getID().equals(me().getPosition())) {
			return Task.minus;
		}
		// 到達したRefugeは無視
		if (visitedArea.contains(area)) {
			return Task.minus;
		}
		// 到達可能なら無視
		if (reachable.getReachableType(area).equals(ReachableType.True)) {
			return Task.minus;
		}
		// 行き止まりは無視
		if (area instanceof Road && area.getNeighbours().size() < 2) {
			return Task.minus;
		}
		if (area instanceof Building 
				&& ((Building)area).isFierynessDefined()
				&& ((Building)area).getFierynessEnum().equals(Fieryness.BURNT_OUT)) {
			return Task.minus;
		}

		double result = Task.low;

		if (area.isBlockadesDefined()) {
			List<EntityID> blocks = area.getBlockades();
			// 必要なBlockadeを数える
			int count = 0;
			for (EntityID blockID : blocks) {
				Blockade b = (Blockade) world.getEntity(blockID);
				if (!unnecessaryBlockades.contains(b)) {
					count++;
				}
			}
			if (count == 0) {
				return Task.minus;
			}
			else {
				result = Task.high;
			}
		}
		if (!reachable.getReachableType(area).equals(ReachableType.True)) {
			result = Task.high;
		}
		else if (area instanceof Building) {
			if (world.maybeBurnedOut.contains(area)) {
				result /= 3.0;
			}
			else {
				for (EntityID id : world.needRescueHuman) {
					Human hm = (Human) world.getEntity(id);
					if (hm != null && hm.isPositionDefined() && area.equals(hm.getPosition(world))) {
						result = Task.high;
					}
				}
			}
		}
		if (someoneVisitedArea.contains(area)) {
			result -= Task.middle;
		}

		// 周辺のエージェント考慮
		result *= aroundAgentsLate(area, (viewDistance * time / timestep) >> 2);// 時間が進むごとに広げてみる
		// 周辺の火災考慮
		result *= aroundFireLate(area);
		// 距離調整
		result *= distanceLate(area);

		return (int) result;
	}

	private int getHumanPriority(Human human) {
		Area area = null;
		// PFは無視
		if (human instanceof PoliceForce) {
			return Task.minus;
		}
		if (human.isPositionDefined()) {
			try {
				area = (Area) world.getEntity(human.getPosition());
			}
			catch (ClassCastException e) {
				// PositionがATの場合は無視
				return Task.minus;
			}
			if (!world.isLocateInBlockade(human)) {
				// 到達したことのある場所に居る人は無視
				if (visitedArea.contains(area)) {
					return Task.minus;
				}
				// たぶん助かった人も無視
				// if (isMaybeSurvivor(human)) {
				// return Task.minus;
				// }
				if (reachable.getReachableType(area).equals(ReachableType.True)) {
					// 到達可能な人も無視
					return Task.minus;
				}
			}
		}
		else {
			return Task.minus;
		}
		if (human.getPosition(world) == null) {
			return Task.minus;
		}
		if (human.isDamageDefined() && human.isHPDefined()) {
			if (world.getDeathTime(human) < human.getBuriedness() / time + 4) {
				// 死んだ・たぶん死ぬ人も無視
				return Task.minus;
			}
		}

		double result = Task.middle;
		if (world.needRescueHuman.contains(human.getID())) {
			// 救助すべきひと
			result = Task.high;
		}

		// 周辺のエージェント考慮
		// result *= aroundAgentsLate(human,
		// (viewDistance * time / timestep) >> 2);//時間が進むごとに広げてみる
		// 距離調整
		result *= distanceLate(human);
		// 種類調整
		result *= humanTypeLate(human);

		return (int) result;
	}

	/**
	 * Humanの種類と優先度の関係
	 * 
	 * @param human
	 * @return エージェントの種類で考慮される優先度の重み(0.0〜1.0)
	 */
	private double humanTypeLate(Human human) {
		// そのときのチームのバランスで決め打ち
		if (human instanceof AmbulanceTeam) {
			return 0.8;
		}
		else if (human instanceof Civilian) {
			if (time < timestep / 2) {
				// 前半
				return 0.1;
			}
			else {
				// 後半
				return 0.8;
			}
		}
		else if (human instanceof FireBrigade) {
			if (time < timestep / 2) {
				// 前半
				return 0.9;
			}
			else {
				// 後半
				return 0.7;
			}
		}
		else if (human instanceof PoliceForce) {
			return 0.4;
		}
		return 0.0;
	}

	/**
	 * Entityの距離と優先度の関係
	 * 
	 * @param entity
	 *                対象のEntity
	 * @return 距離で考慮される優先度の重み(0.0〜1.0)
	 */
	private double distanceLate(StandardEntity entity) {
		double distance = world.getDistance(me(), entity);
		double result = 0.0;

		if (distance < 0.0) {
			// isPositionDefined() == falseの場合
			if (time < timestep / 3) {
				result = 0.2;
			}
			else {
				// 後半で知らないものは見たほうがいいかも
				result = 0.8;
			}
		}
		else if (distance < repairDistance) {
			// 除去可能な範囲はどうでもいい
			result = 0.0;
		}
		else if (distance < viewDistance) {
			// 視界内なら情報がはっきりしているので強め
			// 距離が遠いほど高くする
			result = 0.5 + distance / viewDistance;
		}
		else {
			// 見えないけど通信で送られてきた
			if (distance < 2.0 * viewDistance) {
				result = 0.75;
			}
			else {
				result = 0.5;
			}
		}
		return Math.min(1.0, result);
	}

	/**
	 * ターゲットの終了判定
	 * 
	 * @return 終了したらTrue
	 */
	protected boolean checkFinishTarget() {
		// ターゲットが定まっていないなら再考
		if (targetTask == null || targetTask.target == null)
			return true;
		StandardEntity target = world.getEntity(targetTask.target);
		if (target == null)
			return true;
		if (target instanceof Blockade) {
			// 無視できる
			if (unnecessaryBlockades.contains(target)) {
				return true;
			}
			// 自分でなくて良い場合
			if (isClearable((Blockade) target) && !isMytask((Blockade) target)) {
				return true;
			}
		}
		else if (target instanceof Area) {
			// 自分の場所がTargetなら再考
			// 誰かが自分の行くべきAreaに先に到達したら再考
			return world.getEntity(me().getPosition()).equals(target) || passedRefuge.contains(target) || someoneVisitedArea.contains(target);
		}
		else if (target instanceof Human) {
			if (((Human) target).isPositionDefined()) {
				return me().getPosition().equals(((Human) target).getPosition()) || isMaybeSurvivor((Human) target) || !world.isLocateInBlockade((Human) target);
			}
		}
		return false;
	}

	@Override
	// 仕事があってもrandom walkに入るときがある. おそらくはisMyTaskの使い方があまい
	protected void act() throws ActionCommandException {
		// moveしようとしているにも関わらず移動できていないとき
		// 建物から出られないときの対策
		try {
			if (getCommandHistory(time - 1).equals(StandardMessageURN.AK_MOVE) && getCommandHistory(time - 2).equals(StandardMessageURN.AK_MOVE)) {
				if (getLocationHistory(time - 1).equals(getLocationHistory(time - 2))) {
					switch (time % 2) {
					case 0:
						if (location() instanceof Building) {
							Building bld = (Building) location();
							goOut(bld);
						}
						break;
					case 1:
						clearNear();
						break;
					}
				}
			}
		}
		catch (NullPointerException e) {
		}

		clearAround();

		// ダメージを受けたとき
		if (me().getHP() - me().getDamage() * (timestep - time) < 0) {
			moveToRefuge();
		}

		moveToTarget();

		searchBuildings();

		randomWalk();
	}

	/**
	 * 自分のいるエリア内で移動せずに瓦礫除去可能な範囲を除去する．<br>
	 * 自分のいる位置から移動せずに隣接エリア内に瓦礫が除去できる場合も除去する．
	 */
	protected void clearAround() throws ActionCommandException {
		ArrayList<Blockade> blocks = new ArrayList<Blockade>();
		for (EntityID id : getVisibleEntitiesID()) {
			StandardEntity entity = world.getEntity(id);
			if (entity instanceof Blockade) {
				Blockade block = (Blockade) entity;
				if (unnecessaryBlockades.contains(block)) {
					continue;
				}
				if (!isClearable(block)) {
					continue;
				}
				// 到達可能なRefugeが無い場合は，Refuge以外の建物を塞ぐ瓦礫を無視
				if (passedRefuge.size() == 0 && reachable.isIgnorableBlockad(block, false)) {
					continue;
				}
				if (!isMytask(block)) continue;
				blocks.add(block);
			}
		}
		if (blocks.size() == 0)
			return;
		Collections.sort(blocks, new DistanceSorter(me(), world));
		clear(blocks.get(time % blocks.size()));
	}

	private void clearNear() throws ActionCommandException {
		ArrayList<Blockade> blocks = new ArrayList<Blockade>();
		for (EntityID id : getVisibleEntitiesID()) {
			StandardEntity entity = world.getEntity(id);
			if (entity instanceof Blockade) {
				Blockade block = (Blockade) entity;
				blocks.add(block);
			}
		}
		Collections.sort(blocks, new DistanceSorter(me(), world));

		for (Blockade block : blocks) {
			if (!isClearable(block)) {
				continue;
			}
			clear(block);
		}
	}

	/**
	 * id(BlockadeのEntityID)が自分が除去すべき Taskかどうかを判定するのに使う
	 * 
	 * @param block
	 * @return 自分のTaskと判定されたときtrue
	 */
	private boolean isMytask(Blockade block) {
		// 情報が不明の場合は判断しない
		if (!block.isRepairCostDefined()) {
			return true;
		}
		// blockをClearできるPoliceForceのリスト
		ArrayList<StandardEntity> nearPoliceForces = new ArrayList<StandardEntity>();
		for (StandardEntity next : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
			if (isVisible(next)) {
				PoliceForce pf = (PoliceForce) next;
				if (!isClearable(pf, block)) {
					continue;
				}

				/*
				 * 自分が建物内にいる際には, 建物外にいるエージェントと同時にClearをして
				 * サーバが落ちないようにするために, Clearを行わないようにしている.
				 * サーバ側が落ちないようになったら, 外してよい.
				 * サーバ側が修正されたのでコメントアウトしました.
				 */
				// StandardEntity entity =
				// world.getEntity(pf.getPosition());
				// if(entity instanceof Building){continue;}

				nearPoliceForces.add(next);
			}
		}

		/*
		 * 自分が建物内にいる際には, 建物外にいるエージェントと同時にClearをして サーバが落ちないようにするために,
		 * Clearを行わないようにしている. サーバ側が落ちないようになったら, 外してよい.
		 * サーバ側が修正されたのでコメントアウトしました.
		 */
		// StandardEntity entity = world.getEntity(me().getPosition());
		// if(entity instanceof Building){
		// return false;
		// }

		// System.out.println(getID()+ " " +nearPoliceForces);
		Collections.sort(nearPoliceForces, new IdSorter());
		if (!block.isRepairCostDefined())
			return true;

		/*
		 * blockadeを除去するのに必要な最大の人数 ただし, 10を超える,
		 * RepairCostを複数人で一度にClearすると サーバが落ちるので,
		 * RepairCostが10以上の時はmax未満の人数で Clearをする必要がある.
		 */
		double max = Math.max(Math.ceil((float) block.getRepairCost() / repairRate) - 1.0, 1.0);
		return nearPoliceForces.indexOf(world.getEntity(me().getID())) < max;

		/*
		 * 現在使用していない, return
		 * nearPoliceForces.indexOf(world.getEntity(me().getID())) < max;
		 * がおかしいとkの 保険として残してある.
		 */
		// if(nearPoliceForces.size() * repairRate <
		// block.getRepairCost()){
		// return true;
		// }
		// return
		// world.getUniform(me().getID())==world.getUniform(nearPoliceForces.get(0).getID());

	}

	protected void moveToTarget() throws ActionCommandException {
		if (targetTask == null || targetTask.target == null) {
			return;
		}
		switch (targetTask.approach) {
		case PoliceForceTask.PF_FUNCTION:
			// 瓦礫を無視した最低コスト
			move(targetTask.target, CostFunction.pf(reachable, world));
			break;
		default:
			// 最低コスト
			move(targetTask.target);
			break;
		}
	}

	/**
	 * おそらく助かるであろうと思われる人
	 * 
	 * @param human
	 * @return
	 */
	private boolean isMaybeSurvivor(Human human) {
		if (!human.isPositionDefined())
			return false;
		StandardEntity area = world.getEntity(human.getPosition());
		// 救助済み
		if (area instanceof Refuge) {
			return true;
		}
		// 搬送中
		if (area instanceof AmbulanceTeam) {
			return true;
		}
		// 路上
		if (area instanceof Road) {
			return true;
		}
		return false;
	}

	@Override
	public String toString() {
		return "Ri-one police force agent";
	}
}
