
package rione.communication;

import rescuecore2.config.Config;
import rescuecore2.standard.entities.StandardEntityURN;
import rione.agent.AgentConstants;
import rione.object.AdvancedWorldModel;

public abstract class MessageConstants {

	private static final String HP_PERSEPTION_KEY = "perception.los.precision.hp";
	private static final String DAMAGE_PERSEPTION_KEY = "perception.los.precision.damage";

	private static final int DEFAULT_HP_PERSEPTION = 1000;
	private static final int DEFAULT_DAMAGE_PERSEPTION = 100;

	protected final AdvancedWorldModel world;

	/** 通信に使用するオブジェクトのヘッダの列挙． */
    public static final int
    HEADER_ROAD             = 0,
    HEADER_BUILDING         = 1,
    HEADER_BLOCKADE         = 2,
    HEADER_CIVILIAN         = 3,
    HEADER_PLATOON          = 4,
    HEADER_TASK             = 5,
    HEADER_KIND             = 6; //何種類かを必ず書くこと
    /** オブジェクトヘッダのサイズ． */
    public static final int OBJECT_HEADER_BIT_SIZE  = CommunicationUtil.needBitSize(HEADER_KIND);

    /** platoonのヘッダの列挙． */
    protected static final int
    PLATOON_TYPE_FB	= 0,
    PLATOON_TYPE_PF	= 1,
    PLATOON_TYPE_AT	= 2,
    PLATOON_TYPE_KIND = 3;
    /** Platoonヘッダのサイズ */
	protected static final int PLATOON_HEADER_BIT_SIZE = CommunicationUtil.needBitSize(PLATOON_TYPE_KIND);

	/** オブジェクト識別子(ID,Uniform)用 */
    private final int BUILDINGS_SIZE;
    private final int ROADS_SIZE;
	private final int AT_SIZE;
	private final int FB_SIZE;
	private final int PF_SIZE;
	protected final int BUILDINGS_BIT_SIZE;
	protected final int ROADS_BIT_SIZE;
	protected final int AT_UNIFORM_BIT_SIZE;
	protected final int FB_UNIFORM_BIT_SIZE;
	protected final int PF_UNIFORM_BIT_SIZE;
	protected final int CIVILIAN_ID_BIT_SIZE = 32;
	protected final int BlOCKADE_ID_BIT_SIZE = 32;
	protected final int ENTITY_ID_BIT_SIZE = 32;

	/** 座標表現用 */
	protected final int MAX_X;
	protected final int MIN_X;
	protected final int MAX_Y;
	protected final int MIN_Y;
	protected final int X_BIT_SIZE;
	protected final int Y_BIT_SIZE;

    /** Roadの書き込み用 */
	protected final int
	ROAD_PASSABLE_BIT_SIZE			= 1;

    /** Buildingの書き込み用 */
    protected final int
	BUILDING_FIERYNESS_BIT_SIZE		= 3;

    /** Blockadeの書き込み用 */
    protected final int
    BLOCKADE_REPAIR_COST_BIT_SIZE	= 10;

    /** Human(Civilian,Platoon)の書き込み用 */
    protected final int
    HUMAN_HP_BIT_SIZE,
    HUMAN_BURIEDNESS_BIT_SIZE,
    HUMAN_DAMAGE_BIT_SIZE;

	protected final int HP_PERSEPTION;
	protected final int DAMAGE_PERSEPTION;

    /** Civilianの書き込み用 */
    //none

    /** Platoonの書き込み用 */
    protected final int
	PROPERTY_FB_WATER_BIT_SIZE,
	PLATOON_IS_LOCATE_IN_BLOCKADE_BIT_SIZE = 1;

    /** 送信するオブジェクトのサイズ */
	public final int
    ROAD_MESSAGE_BIT_SIZE,
    BUILDING_MESSAGE_BIT_SIZE,
    CIVILIAN_MESSAGE_BIT_SIZE,
    BLOCKADE_MESSAGE_BIT_SIZE,
    FB_MESSAGE_BIT_SIZE,
    PF_MESSAGE_BIT_SIZE,
    AT_MESSAGE_BIT_SIZE;

	public MessageConstants(AdvancedWorldModel world, Config config) {
		this.world = world;

		//オブジェクト識別子(ID,Uniform)の初期化
		this.AT_SIZE = world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM).size();
		this.AT_UNIFORM_BIT_SIZE = CommunicationUtil.needBitSize(AT_SIZE);
		this.FB_SIZE = world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE).size();
		this.FB_UNIFORM_BIT_SIZE = CommunicationUtil.needBitSize(FB_SIZE);
		this.PF_SIZE = world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE).size();
		this.PF_UNIFORM_BIT_SIZE = CommunicationUtil.needBitSize(PF_SIZE);

		//座標表現用の初期化
		this.MIN_X = world.getWorldBounds().first().first();
		this.MIN_Y = world.getWorldBounds().first().second();
		this.MAX_X = world.getWorldBounds().second().first();
		this.MAX_Y = world.getWorldBounds().second().second();
		this.X_BIT_SIZE = CommunicationUtil.needBitSize(MAX_X-MIN_X);
		this.Y_BIT_SIZE = CommunicationUtil.needBitSize(MAX_Y-MIN_Y);

		//Road用の初期化
		this.ROADS_SIZE = world.getEntitiesOfType(StandardEntityURN.ROAD).size();
		this.ROADS_BIT_SIZE = CommunicationUtil.needBitSize(ROADS_SIZE);
		ROAD_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + ROADS_BIT_SIZE;

		//Building用の初期化
		this.BUILDINGS_SIZE = world.getEntitiesOfType(AgentConstants.BUILDINGS).size();
		this.BUILDINGS_BIT_SIZE = CommunicationUtil.needBitSize(BUILDINGS_SIZE);
		BUILDING_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + BUILDINGS_BIT_SIZE + BUILDING_FIERYNESS_BIT_SIZE;

		//Blockade用の初期化
		BLOCKADE_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + BlOCKADE_ID_BIT_SIZE + X_BIT_SIZE + Y_BIT_SIZE;

		//Human用の初期化
		this.HP_PERSEPTION = config.getIntValue(HP_PERSEPTION_KEY, DEFAULT_HP_PERSEPTION);
		this.HUMAN_HP_BIT_SIZE			= CommunicationUtil.needBitSize(10000/HP_PERSEPTION);
		this.DAMAGE_PERSEPTION = config.getIntValue(DAMAGE_PERSEPTION_KEY, DEFAULT_DAMAGE_PERSEPTION);
		this.HUMAN_DAMAGE_BIT_SIZE		= CommunicationUtil.needBitSize(2000/DAMAGE_PERSEPTION);
		this.HUMAN_BURIEDNESS_BIT_SIZE	= 7;
		int humanBitSize = HUMAN_HP_BIT_SIZE + HUMAN_DAMAGE_BIT_SIZE+ HUMAN_BURIEDNESS_BIT_SIZE + X_BIT_SIZE + Y_BIT_SIZE;

		//Civilianの初期化
	    CIVILIAN_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + CIVILIAN_ID_BIT_SIZE + humanBitSize;

		//Platoon用の初期化
		PROPERTY_FB_WATER_BIT_SIZE = CommunicationUtil.needBitSize(config.getIntValue("resq-fire.water_capacity", 15000)/100);
	    FB_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + PLATOON_HEADER_BIT_SIZE + FB_UNIFORM_BIT_SIZE + PROPERTY_FB_WATER_BIT_SIZE + humanBitSize;
	    PF_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + PLATOON_HEADER_BIT_SIZE + PF_UNIFORM_BIT_SIZE + humanBitSize;
	    AT_MESSAGE_BIT_SIZE = OBJECT_HEADER_BIT_SIZE + PLATOON_HEADER_BIT_SIZE + AT_UNIFORM_BIT_SIZE + humanBitSize;

	}
}
