package rione.communication;

import java.util.PriorityQueue;

import rescuecore2.config.Config;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.CentreAgent;
import rione.agent.CommunicationAgent;
import rione.communication.channel.AbstractChannel;
import rione.communication.channel.RadioChannel;
import rione.communication.channel.VoiceChannel;
import rione.object.AdvancedWorldModel;

public class MessageManager {

	CommunicationAgent<?> agent;

	private final AdvancedWorldModel world;
	private final Config config;

	/** 使用可能な帯域幅 */
	private int bandwidthLimit;

	/** 送信するbit列をByte配列で書きだすライタ */
	private MessageWriter writer;

	/** 送信情報を格納する優先度付きキュー */
	private PriorityQueue<MessageData> radioMessages;
	/** 送信情報を格納する優先度付きキュー */
	private PriorityQueue<MessageData> voiceMessages;

	public MessageManager(CommunicationAgent<?> agent, AdvancedWorldModel world, Config config) {
		this.agent = agent;
		this.world = world;
		this.config = config;
		bandwidthLimit = (int)(1.5 * agent.getChannelManager().getBandwidthAll() / Math.max(world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.AMBULANCE_TEAM, StandardEntityURN.POLICE_FORCE).size(), 1));
		writer = new MessageWriter(world, config);
		radioMessages = new PriorityQueue<MessageData>();
		voiceMessages = new PriorityQueue<MessageData>();
	}

	public void init() {
		radioMessages.clear();
		voiceMessages.clear();
	}

	public void addSendData(StandardEntity entity) {
		AbstractMessageMaker maker = new RadioMessageMaker(world);
		radioMessages.add(maker.make(entity));
		maker = new VoiceMessageMaker(world);
		voiceMessages.add(maker.make(entity));
		maker = null;
	}
	public void addSendData(EntityID id) {
		addSendData(world.getEntity(id));
	}

	public byte[] getData(int c) {
		writer.init();
		MessageWriter tester = new MessageWriter(world, config);
		//------------------------------------------------
		//------------------------------------------------
		PriorityQueue<MessageData> dataQueue;
		//チャネル番号からチャネル情報を取得
		AbstractChannel channel = agent.getChannelManager().getChannel(c);
		//チャネル情報から無線か声かを判断
		if (channel instanceof RadioChannel) {
			//無線の場合
			dataQueue = new PriorityQueue<MessageData>(radioMessages);
			bandwidthLimit = (int)(1.5 * agent.getChannelManager().getBandwidthAll() / Math.max(world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.AMBULANCE_TEAM, StandardEntityURN.POLICE_FORCE).size(), 1));
		} else {
			//声の場合
			dataQueue = new PriorityQueue<MessageData>(voiceMessages);
			bandwidthLimit = ((VoiceChannel)channel).getSize();
		}

		//送信情報の書き出し
//		String str = "[";
		while (!dataQueue.isEmpty()) {
			MessageData message = dataQueue.poll();
			if (message instanceof StanderdEntityMessageData && !(agent instanceof CentreAgent<?>))
			{
				StandardEntity entity = ((StanderdEntityMessageData) message).getEntity();
				//送信する上限または優先度が負の値になったらbyte配列の書き込みを終わる
				tester.write(entity);
	//			str += entity + ",";
				if (bandwidthLimit <= tester.getDatasize()) {
					break;
				}
				writer.write(entity);
			}
		}
//		System.out.println(world.time+","+c+","+writer.getData().length+","+str+"]");

		tester = null;
		dataQueue = null;
		return writer.getData();
	}

}
