package rione.communication;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rione.object.AdvancedWorldModel;
import rione.object.ReachableArea;

public class RadioMessageMaker extends AbstractMessageMaker {

	public RadioMessageMaker(AdvancedWorldModel world) {
		super(world);
	}

	@Override
	public MessageData make(StandardEntity entity) {
		StanderdEntityMessageData data = new StanderdEntityMessageData();
		int priority = 0;

		data.setEntity(entity);
		//--------------------
		// 優先度評価関数作る;
		//通信のpriorityのdefaltの値は0
		//--------------------
		switch (entity.getStandardURN()) {
		case ROAD :
			priority = -1;
			if (world.me() instanceof PoliceForce) {
				ReachableArea reachable = world.getReachable();
				if (reachable.isRecognized((Road)entity)
						&& reachable.isPassable((Road)entity)) {
					priority = 200;
				}
			}
			break;
		//case BLOCKADE :
		//	break;
		case BUILDING :
			priority = 100;
			if (((Building)entity).getFierynessEnum() == Fieryness.UNBURNT) {
				priority = 50;
			}
			if (((Building)entity).getFierynessEnum() == Fieryness.HEATING) {
				priority = 120;
			}
			break;
		case CIVILIAN :
			Civilian cv = (Civilian) entity;
			if (cv.isPositionDefined() && !world.needRescueHuman.contains(cv.getID())) {
				// 助ける必要のないCivilian
				priority = Integer.MIN_VALUE;
			}
			else if (cv.isPositionDefined() && world.getEntity(cv.getPosition()) instanceof AmbulanceTeam) {
			    priority = 110;
			}
			else {
				int deathTime = world.getDeathTime(cv);
				if (deathTime <= world.time + 5 || (cv.isHPDefined() && cv.getHP() <= 0)) {
					//死んでるなら送らない(最善の場合なので少し増やしている)
					priority = Integer.MIN_VALUE;
				}
				else if (world.timestep + 50 < deathTime) {
					//死にそうにない
					priority = Integer.MIN_VALUE;
				}
				else if (world.time % 40 == 39) {
					//40サイクルに1回高くする
					priority = 120;
				}
				else if (2 * deathTime - world.time < world.timestep) {
					// deathTime - time < (timestep - time) / 2からの変形
					// timestep以内に死にそうな人前半
					priority = 60;
				}
				else {
					priority = 30;
				}
			}
			break;
		case FIRE_BRIGADE :
			if(world.me() != null && world.me().getID().equals(entity.getID())){
				if( ((FireBrigade) world.me()).getWater() == 0
						&& (world.getEntity(world.me().getPosition()) instanceof Refuge)
				){
					priority = 199;
					break;
				}
				else{priority = 20; break;}
			}
			priority = 9;
			break;
		case POLICE_FORCE :
				if(world.time <= 3 &&world.me() != null && world.me().getID().equals(entity.getID())){
			priority = 10;
			break;
				}
				priority = 5;
				break;
		case AMBULANCE_TEAM :
				if(world.time <= 3 &&world.me() != null && world.me().getID().equals(entity.getID())){
			priority = 10;
			break;
				}
				priority = 5;
				break;
			//		case REFUGE :
			//			break;
		default :
			priority = -1;
			break;
		}
		data.setPriotiry(priority);
		return data;
	}


}
