package rione.communication.channel;

import java.util.ArrayList;

public abstract class AbstractChannel {

	private final int channel;
	
	private ArrayList<AbstractNoise> inputNoise;
	
	private ArrayList<AbstractNoise> outputNoise;
	
	public AbstractChannel(int channel) {
		this.channel = channel;
		inputNoise = new ArrayList<AbstractNoise>();
		outputNoise = new ArrayList<AbstractNoise>();
	}
	
	public int getChannel() {
		return channel;
	}

	/**
	 * 通信した際，正しくデータが送られる確率．
	 * @return 正しくデータが送られる確率．
	 */
	public double getReliability() {
		double reliability = 1.0;
		for (AbstractNoise noise: inputNoise) {
			reliability *= 1.0 - noise.getRatio();
		}
		for (AbstractNoise noise: outputNoise) {
			reliability *= 1.0 - noise.getRatio();
		}
		return reliability;
	}

	public boolean addInputNoise(AbstractNoise noise) {
		return inputNoise.add(noise);
	}

	public boolean addOutputNoise(AbstractNoise noise) {
		return outputNoise.add(noise);
	}

	@Override
	public String toString() {
		return "AbstractChannel";
	}
}
