package rione.object;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.AbstractEntity;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

// Taskで扱う便宜上AbstrackEntityを継承
public class Street extends AbstractEntity {

	/** Streetに所属するBuildingのリスト */
	private final ArrayList<Area> buildings = new ArrayList<Area>();
	/** Streetに所属するRoadのリスト */
	private final ArrayList<Area> roads = new ArrayList<Area>();
	/** Streetに所属するすべてのEntity */
	private final Set<Area> entities = new HashSet<Area>();
	/** このStreetのhead */
	private Area head;
	/** このStreetのtail */
	private Area tail;
	/** head側で隣接するStreetのリスト*/
	private final ArrayList<Street> headSideNeighborStreets = new ArrayList<Street>();
	/** tail側で隣接するStreetのリスト*/
	private final ArrayList<Street> tailSideNeighborStreets = new ArrayList<Street>();
	/** あるAreaのhead側のArea
	    あるAreaがBuildingの場合、そのBuildingのentranceが登録される */
	private final Map<Area, Area> heads = new HashMap<Area, Area>(); 
	/** あるAreaのtail側のArea 
	    あるAreaがBuildingの場合、そのBuildingのentranceが登録される */
	private final Map<Area, Area> tails = new HashMap<Area, Area>(); 
	/** Streetを使用した際のRoute */
	private final LinkedList<EntityID> headToTailroute = new LinkedList<EntityID>();
	private final LinkedList<EntityID> tailToHeadroute = new LinkedList<EntityID>();
	
		
    /**
     * コンストラクタ
     * Street内でしか呼び出さない.
     * @param uniform 背番号
     */
	private Street(int uniform){
		super(new EntityID(uniform));
	}
	
	public Street(Street other) {
		super(other);
	}

	public Area getHead() {
		return head;
	}
	
	public Area getTail() {
		return tail;
	}
	
	public int getX() {
		return 0;
	}
	
	public int getY() {
		return 0;
	}
	
	public ArrayList<Street> getHeadSideNeignbhorStreets() {
		return headSideNeighborStreets;
	}
	
	public ArrayList<Street> getTailSideNeignbhorStreets() {
		return tailSideNeighborStreets;
	}
	
	public List<EntityID> getRouteFrom(Area area) {
		if (area.equals(head)) {
			return headToTailroute;
		} else if (area.equals(tail)) {
			return tailToHeadroute;
		}
		return null; // ここには到達しないように呼び出す
	}
	
	public List<EntityID> getRouteTo(Area area) {
		if (area.equals(head)) {
			return tailToHeadroute;
		} else if (area.equals(tail)) {
			return headToTailroute;
		}
		return null; // ここには到達しないように呼び出す
	}
	
	public List<EntityID> getTailToHeadRoute() {
		return headToTailroute;
	}

	public List<EntityID> getHeadToTailRoute() {
		return tailToHeadroute;
	}
	
	public Area getOppositeArea(Area area) {
		if (area.equals(head)) {
			return tail;
		} else if (area.equals(tail)) {
			return head;
		}
		return null;
	}
	
	public List<Street> getOppositeNeightbor(Area area) {
		if (area.equals(head)) {
			return tailSideNeighborStreets;
		} else if (area.equals(tail)) {
			return headSideNeighborStreets;
		}
		return null;
	}
	
	/**
	 * Streetを作成する.
	 * @param world WorldModel
	 * @return Streetのリスト
	 */
	protected static ArrayList<Street> createStreet(AdvancedWorldModel world) {
		int new_uniform = 1;
		
		Set<StandardEntity> notIsolationEntities = new HashSet<StandardEntity>();
		ArrayList<Street> streets = new ArrayList<Street>();
		ArrayList<StandardEntity> roads = new ArrayList<StandardEntity>(world.getEntitiesOfType(StandardEntityURN.ROAD));
		while (!roads.isEmpty()) {
			Road road = (Road) roads.remove(0);
			if (!world.isStreetRoad(road)) {
				continue;
			}
			List<EntityID> neighborhood = road.getNeighbours();
			ArrayList<Area> neighborhoodBuildings = new ArrayList<Area>();
			for (EntityID id : neighborhood) {
				StandardEntity entity = world.getEntity(id);
				if (entity instanceof Building){
					neighborhoodBuildings.add((Building)entity);
				}
			}
			while (!neighborhood.isEmpty()) {
				StandardEntity entity = world.getEntity(neighborhood.remove(0));
				if (entity instanceof Road) {
					Street street = new Street(new_uniform++);
					street.head = road;
					world.setOwnStreet(road.getID(), street);
					street.buildings.addAll(neighborhoodBuildings);	
					street.entities.addAll(neighborhoodBuildings);	
					for (Area area : neighborhoodBuildings) {
						street.heads.put(area, road);
						street.tails.put(area, road);
					}
					street.entities.add((Road)entity);
					world.setOwnStreet(entity.getID(), street);
					street.roads.add((Road)entity);
					street.headToTailroute.add(entity.getID());
					street.tailToHeadroute.addFirst(entity.getID());
					street.heads.put((Road)entity, road);
					street.tails.put(road, (Road)entity);
					expandStreet(street, (Road)entity, world);
					if (!streets.contains(street)) {
						streets.add(street);
						notIsolationEntities.addAll(street.entities);
					}
				}
			}
		}
		world.setIsolationEntities(notIsolationEntities);
		setNeighborStreet(streets);
		return streets;
	}

	/**
	 * 再帰的にStreetを拡張する.
	 */
	private static void expandStreet(Street street, Road road, AdvancedWorldModel world) {
		List<EntityID> neighborhood = road.getNeighbours();
		while (!neighborhood.isEmpty()) {
			Area entity = (Area)world.getEntity(neighborhood.remove(0));
			if (street.entities.contains(entity)) { 
				continue; 
			}
			street.entities.add(entity);
			world.setOwnStreet(entity.getID(), street);
			if (entity instanceof Building) {
				street.buildings.add(entity);
				street.heads.put(entity, road);
				street.tails.put(entity, road);
			}
			else if (entity instanceof Road) {
				street.roads.add(entity);
				street.headToTailroute.add(entity.getID());
				street.tailToHeadroute.addFirst(entity.getID());
				street.heads.put(entity, road);
				street.tails.put(road, entity);
				if (world.isStreetRoad((Road)entity)) {
					street.tail = entity;
				} else {
					expandStreet(street, (Road)entity, world);
				}
			}
		}
	}
	
	/**
	 * Streetに隣接するStreetを設定します.
	 * @param streets 全Streetのリスト
	 */
	private static void setNeighborStreet(List<Street> streets) {
		for (int i=0; i<streets.size()-1; i++) {
			Street street1 = streets.get(i);
			for (int j=i+1; j<streets.size(); j++) {
				Street street2 = streets.get(j);
				if (street1.head.equals(street2.head)) {
					street1.headSideNeighborStreets.add(street2);
					street2.headSideNeighborStreets.add(street1);
				} else if (street1.head.equals(street2.tail)) {
					street1.headSideNeighborStreets.add(street2);
					street2.tailSideNeighborStreets.add(street1);
				}
				if (street1.tail.equals(street2.head)) {
					street1.tailSideNeighborStreets.add(street2);
					street2.headSideNeighborStreets.add(street1);
				} else if (street1.tail.equals(street2.tail)) {
					street1.tailSideNeighborStreets.add(street2);
					street2.tailSideNeighborStreets.add(street1);
				}
			}
		}
	}
	
	/**
	 * RoutaPlan生成のための一時的なStreetを返す.
	 * 返されたStreetは、RoutePlanに関する一部の機能しか使用できない.
	 * @param from　分割するStreetの始点
	 * @param to    分割するStreetの終点
	 * @return RoutePlan生成のために分割されたStreet
	 */
	public Street divideStreet(Area from, Area to){
		//TODO moveの際に常に使用するので、計算時間について考慮する
		Street street = new Street(-1);
		street.head = from;
		street.headToTailroute.add(from.getID());
		
		// Streetの端が to だった場合、新しいStreetの tail と tailSideNeignborStreet を登録する 
		if (head.equals(to)) {
			street.tail=to;
			street.headSideNeighborStreets.addAll(this.headSideNeighborStreets);
		} else if (tail.equals(to)) {
			street.tail=to;
			street.tailSideNeighborStreets.addAll(this.tailSideNeighborStreets);
		}
		
		LinkedList<EntityID> route = new LinkedList<EntityID>();
		route.add(from.getID());
		// tail側に辿る
		for (Area next=tails.get(from); next!=null; next=tails.get(next)) {
			route.add(next.getID());
			if (next.getNeighbours().contains(to.getID())) {
				route.add(to.getID());
				street.headToTailroute.addAll(route);
				return street;
			}
		}
		route.clear();
		route.add(from.getID());
		// head側に辿る
		for (Area next=heads.get(from); next!=null; next=heads.get(next)) {
			route.add(next.getID());
			if (next.getNeighbours().contains(to.getID())) {
				route.add(to.getID());
				street.headToTailroute.addAll(route);
				return street;
			}
		}
		// ここには到達しないはず
		return null;
	}
	
	public List<EntityID> divideRoute(Area from, Area to) {
		List<EntityID> route = new ArrayList<EntityID>();
		route.add(from.getID());
		// tail側に辿る
		for (Area next=tails.get(from); next!=null; next=tails.get(next)) {
			route.add(next.getID());
			if (next.equals(to)) {
				return route;
			}
		}
		route.clear();
		route.add(from.getID());
		// tail側に辿る
		for (Area next=heads.get(from); next!=null; next=heads.get(next)) {
			route.add(next.getID());
			if (next.equals(to)) {
				return route;
			}
		}
		System.out.println("ERROR: ");
		// ここには到達しないはず
		return null;
	}
	
	public boolean contains(StandardEntity entity) {
		return this.entities.contains(entity);
	}
	
	@Override
	public boolean equals(Object obj){
		Street street = (Street) obj;
		return ( this.head.equals(street.head) && this.tail.equals(street.tail) )
			 || ( this.head.equals(street.tail) && this.tail.equals(street.head) );
	}
	
	@Override
	public int hashCode(){
		return head.getID().getValue() * tail.getID().getValue();
	}
	
	@Override
	protected Entity copyImpl() {
		return new Street(this);
	}
	
	@Override
	public String getURN() {
		return RioneEntityURN.STREET.toString();
	}
}
