package rione.task;

import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

public abstract class Task implements Comparable<Task> {

	/**
	 * Taskの対象
	 */
	public final EntityID target;

	/**
	 * タスクへの取り組み方を示す値。<br>
	 * object+approachでどのような方法で何を行うのかを表すことができる。<br>
	 * 具体的な使い方は各派生先で．
	 */
	public int approach;

	/** 現在のTaskの状態(担当者なし、実行中、完了など) */
	public int state;

	/**
	 * タスク処理に必要とされる担当者数。
	 */
	public int needChargeNumber;

	/** タスク評価の優先度 */
	public int priority;
	/** 無視 */
	public static final int minus = -1000;
	/** 微妙（現在は無視） */
	public static final int zero = 0;
	/** Low */
	public static final int low = 1000;
	/** Middle */
	public static final int middle = 5000;
	/** High */
	public static final int high = 10000;

	/**
	 * コンストラクタ
	 * @param entity
	 * 	対象のEntity
	 * @param priority
	 * 	タスクの優先度
	 */
	public Task(StandardEntity entity, int priority) {
		this(entity, -1, 1, priority);
	}
	/**
	 * 	互換性の為に残してあります．<br>
	 * 	コードの簡潔さの面からあまり使用するべきではありません．
	 * @param entity
	 * @param approach
	 * @param needChargeNumber
	 */
	public Task(StandardEntity entity, int approach, int needChargeNumber) {
		this(entity, Task.zero, approach, needChargeNumber);
	}
	/**
	 * コンストラクタ
	 * @param entity
	 * @param approach
	 * @param needChargeNumber
	 * @param priority
	 */
	public Task(StandardEntity entity, int priority, int approach, int needChargeNumber) {
		this.target = entity.getID();
		this.approach = approach;
		this.needChargeNumber = needChargeNumber;
		this.priority = priority;
	}

	@Override
	public boolean equals(Object o) {
		if (!(o instanceof Task)) return false;
		Task t = (Task) o;
		return this.target.equals(t.target) && this.approach == t.approach;
	}

	@Override
	public int hashCode() {
		return target.getValue();
	}

	@Override
	public int compareTo(Task t) {
		return t.priority - this.priority;
	}
}
