package rione.util;

import java.util.Comparator;

import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;

/**
   A comparator that sorts entities by distance to a reference point.
*/
public class DistanceSorter implements Comparator<StandardEntity> {
    
	private final StandardEntity reference;
    private final StandardWorldModel world;
	private final OrderBy order;
	
    /**
       Create a DistanceSorter.
       @param reference The reference point to measure distances from.
       @param world The world model.
    */
    public DistanceSorter(StandardEntity reference, StandardWorldModel world, OrderBy order) {
    	this.reference = reference;
    	this.world = world;
    	this.order = order;
    }
    
    public DistanceSorter(StandardEntity reference, StandardWorldModel world) {
        this(reference, world, OrderBy.ASC);
    }
    
    @Override
    public int compare(StandardEntity a, StandardEntity b) {
        int d1 = world.getDistance(reference, a);
        int d2 = world.getDistance(reference, b);
        return (d1 - d2) * order.by();
    }
}

